#!/usr/bin/env python
# -*- coding: ASCII -*-

"""

:Author: Martin Kircher
:Contact: mkircher@uw.edu
:Date: *20.10.2014
"""

import sys, os
ROOTdir = os.path.dirname(os.path.realpath(__file__))+"/../../"
sys.path.append(os.path.dirname(os.path.realpath(__file__)))
ROOTanno = (ROOTdir+"/annotations").replace("//",'/')

from optparse import OptionParser
from collections import defaultdict
import pysam
import string
import random
import gzip
from bx.intervals.intersection import Intersecter, Interval
from AnalysisLib import get_from_tabix,get_range_from_tabix,read_genome_fastaindex,get_fasta_mmap,count_GC_CpG,get_DNA
from AnalysisLib import eval_dbSNP_frequencies,eval_ESP_frequencies
from AnalysisLib import emf_init_index,emf_get_base_at_position,format_ancestor

parser = OptionParser("%prog [options]")
parser.add_option("-p", "--PIPE", dest="pipe", help="Read variant effect predictor output from stdin",default=False,action="store_true")

parser.add_option("--vep", dest="vep", help="File with variant effect predictor output (def '')",default="")

parser.add_option("--continuous", dest="continuous", help="Input file is coordinate sorted and does not contain coordinate ranges -- not save for indels! (default OFF)",default=False,action="store_true")

parser.add_option("--bscores", dest="bscores", help="Path to tabix file with b scores (def %s/bscores/bscores.tsv.gz)"%(ROOTanno),default='%s/bscores/bscores.tsv.gz'%(ROOTanno))
  
parser.add_option("--GERPelements", dest="GERPelements", help="Path to tabix file with GERP elements (def %s/gerp/gerp_elements.tsv.gz)"%(ROOTanno),default='%s/gerp/gerp_elements.tsv.gz'%(ROOTanno))
parser.add_option("--GERPscores", dest="GERPscores", help="Path to tabix file with GERP elements (def %s/gerp/gerp_scores.tsv.gz)"%(ROOTanno),default='%s/gerp/gerp_scores.tsv.gz'%(ROOTanno))

parser.add_option("--primatePhyloPscores", dest="primatePhyloPscores", help="Path to tabix file with primate PhyloP conservation scores (def %s/phyloP/primates_nohuman.tsv.gz)"%(ROOTanno),default='%s/phyloP/primates_nohuman.tsv.gz'%(ROOTanno))
parser.add_option("--mammalPhyloPscores", dest="mammalPhyloPscores", help="Path to tabix file with mammal PhyloP conservation scores (def %s/phyloP/placentalMammals_nohuman.tsv.gz)"%(ROOTanno),default='%s/phyloP/placentalMammals_nohuman.tsv.gz'%(ROOTanno))
parser.add_option("--vertebratePhyloPscores", dest="vertebratePhyloPscores", help="Path to tabix file with vertebrate PhyloP conservation scores (def %s/phyloP/vertebrates_nohuman.tsv.gz)"%(ROOTanno),default='%s/phyloP/vertebrates_nohuman.tsv.gz'%(ROOTanno))
parser.add_option("--primatePhastCONSscores", dest="primatePhastCONSscores", help="Path to tabix file with primate PhastCons conservation scores (def %s/phastCons/primates_nohuman.tsv.gz)"%(ROOTanno),default='%s/phastCons/primates_nohuman.tsv.gz'%(ROOTanno))
parser.add_option("--mammalPhastCONSscores", dest="mammalPhastCONSscores", help="Path to tabix file with mammal PhastCons conservation scores (def %s/phastCons/placentalMammals_nohuman.tsv.gz)"%(ROOTanno),default='%s/phastCons/placentalMammals_nohuman.tsv.gz'%(ROOTanno))
parser.add_option("--vertebratePhastCONSPscores", dest="vertebratePhastCONSscores", help="Path to tabix file with vertebrate PhastCons conservation scores (def %s/phastCons/vertebrates_nohuman.tsv.gz)"%(ROOTanno),default='%s/phastCons/vertebrates_nohuman.tsv.gz'%(ROOTanno))

parser.add_option("--mapability20bp", dest="mapability20bp", help="Path to tabix file with mapability scores (def %s/mapability/uniqueness20bpDuke.tsv.gz)"%(ROOTanno),default='%s/mapability/uniqueness20bpDuke.tsv.gz'%(ROOTanno))
parser.add_option("--mapability35bp", dest="mapability35bp", help="Path to tabix file with mapability scores (def %s/mapability/uniqueness35bpDuke.tsv.gz)"%(ROOTanno),default='%s/mapability/uniqueness35bpDuke.tsv.gz'%(ROOTanno))

parser.add_option("--reference", dest="reference", help="Path to reference sequence (fasta index needs to be available) (def %s/reference/reference.fa)"%(ROOTanno),default='%s/reference/reference.fa'%(ROOTanno))
parser.add_option("--transcripts", dest="transcripts", help="Path to Ensembl transcript annotation file (def %s/reference/annotation_v75.gtf.gz)"%(ROOTanno),default='%s/reference/annotation_v75.gtf.gz'%(ROOTanno))
parser.add_option("--exons", dest="exons", help="Path to exome annotation file (def %s/annotation_exons_v75.gtf.gz)"%(ROOTanno),default='%s/reference/annotation_exons_v75.gtf.gz'%(ROOTanno))
parser.add_option("--transcribed", dest="transcribed", help="Path to transcript range annotation file (def %s/annotation_transcripts_v75.tsv.gz)"%(ROOTanno),default='%s/reference/annotation_transcripts_v75.tsv.gz'%(ROOTanno))

parser.add_option("--grantham", dest="grantham", help="Path to Grantham score annotation file (def %s/grantham/grantham_matrix.tsv)"%(ROOTanno),default='%s/grantham/grantham_matrix.tsv'%(ROOTanno))

parser.add_option("--motifs", dest="motifs", help="Path to tabix indexed file of predicted binding motifs (def %s/encodeMotifDB/motifs.tsv.gz)"%(ROOTanno),default='%s/encodeMotifDB/motifs.tsv.gz'%(ROOTanno))

#--------------- START NEW ------------------

parser.add_option("--mirSVR", dest="mirSVR", help="Path to tabix indexed file of predicted microRNA binding motifs (def %s/mirSVR/mirSVR.tsv.gz)"%(ROOTanno),default='%s/mirSVR/mirSVR.tsv.gz'%(ROOTanno))

parser.add_option("--mirTargetScan", dest="mirTargetScan", help="Path to tabix indexed file of predicted TargetScanS seeds (def %s/mirTargetScan/targetScanS.tsv.gz)"%(ROOTanno),default='%s/mirTargetScan/targetScanS.tsv.gz'%(ROOTanno))

parser.add_option("--mutabilityIndex", dest="mutabilityIndex", help="Path to tabix indexed file of mutability index values (def %s/mutabilityIndex/mutability.tsv.gz)"%(ROOTanno),default='%s/mutabilityIndex/mutability.tsv.gz'%(ROOTanno))

parser.add_option("--chromHMM_Epigenomes", dest="chromHMM_Epigenomes", help="Path to tabix indexed file of Epigenome RoadMap chromHMM values (def %s/chromHMM/roadMap_chromHMM.tsv.gz)"%(ROOTanno),default='%s/chromHMM/roadMap_chromHMM.tsv.gz'%(ROOTanno))

parser.add_option("--DNAstructure", dest="DNAstructure", help="Path to tabix indexed DNA secondary structure values (def %s/dnaStructure/dnaStructure.tsv.gz)"%(ROOTanno),default='%s/dnaStructure/dnaStructure.tsv.gz'%(ROOTanno))

parser.add_option("--fitCons", dest="fitCons", help="Path to tabix indexed fitCons values (def %s/fitCons/fitCons_2014-08-21.tsv.gz)"%(ROOTanno),default='%s/fitCons/fitCons_2014-08-21.tsv.gz'%(ROOTanno))

#--------------- END NEW ------------------

parser.add_option("--ESP", dest="ESP", help="Path to ESP VCF file (def %s/ESP/ESP6500SI.vcf.gz)"%(ROOTanno),default='%s/ESP/ESP6500SI.vcf.gz'%(ROOTanno))

parser.add_option("--G1000", dest="G1000", help="Path to 1000G VCF file (def %s/1000G/phase1_v3.20101123.vcf.gz)"%(ROOTanno),default='%s/1000G/phase1_v3.20101123.vcf.gz'%(ROOTanno))
parser.add_option("-f","--filterLowCov", dest="filterLowCov", help="Remove variants only identified from 1000 Genomes Exome sequencing (def OFF)",default=False,action="store_true")

parser.add_option("--segway", dest="segway", help="Path to Segway annotation (def %s/segway/segway.tsv.gz)"%(ROOTanno),default='%s/segway/segway.tsv.gz'%(ROOTanno))

parser.add_option("--H3K27Ac", dest="H3K27Ac", help="Path to Encode H3K27Ac files (def %s/H3K27Ac/tabix_files.lst)"%(ROOTanno),default='%s/H3K27Ac/tabix_files.lst'%(ROOTanno))
parser.add_option("--H3K4Me1", dest="H3K4Me1", help="Path to Encode H3K4Me1 files (def %s/H3K4Me1/tabix_files.lst)"%(ROOTanno),default='%s/H3K4Me1/tabix_files.lst'%(ROOTanno))
parser.add_option("--H3K4Me3", dest="H3K4Me3", help="Path to Encode H3K4Me3 files (def %s/H3K4Me3/tabix_files.lst)"%(ROOTanno),default='%s/H3K4Me3/tabix_files.lst'%(ROOTanno))
parser.add_option("--expression", dest="expression", help="Path to Encode expression files (def %s/expression/tabix_files.lst)"%(ROOTanno),default='%s/expression/tabix_files.lst'%(ROOTanno))
parser.add_option("--nucleosome", dest="nucleosome", help="Path to Encode nucleosome files (def %s/nucleosomes/tabix_files.lst)"%(ROOTanno),default='%s/nucleosomes/tabix_files.lst'%(ROOTanno))
parser.add_option("--openChrom", dest="openChrom", help="Path to Encode OpenChromSynth files (def %s/openChromatin/tabix_files.lst)"%(ROOTanno),default='%s/openChromatin/tabix_files.lst'%(ROOTanno))
parser.add_option("--tfbs", dest="tfbs", help="Path to Encode tfbs track summary file (def %s/tfbs/encodeAwgTfbsUniform.tsv.gz)"%(ROOTanno),default="%s/tfbs/encodeAwgTfbsUniform.tsv.gz"%(ROOTanno))

parser.add_option("--segDups", dest="segDups", help="Path to Segmental Duplication file (def %s/segDups/genomicSuperDups.tsv.gz)"%(ROOTanno),default='%s/segDups/genomicSuperDups.tsv.gz'%(ROOTanno))

parser.add_option("--ancestor_path", dest="ancestor_path", help="PATH of the EMF and EMF index files (def %s/primateWGA)"%(ROOTanno),default="%s/primateWGA"%(ROOTanno))
parser.add_option("--refSpecies", dest="refSpecies", help="Reference species (def Hsap)",default="Hsap")

parser.add_option("-a","--all", dest="all", help="Produce all output lines, rather than random one on same hierarchy (def OFF)",default=False,action="store_true")
parser.add_option("-r","--random", dest="random", help="Produce a completely random line, rather than random one on same hierarchy (def OFF)",default=False,action="store_true")
parser.add_option("--noheader", dest="noheader", help="Turn off header line (def OFF)",default=False,action="store_true")
parser.add_option("-v","--verbose", dest="verbose", help="Turn verbose messages on (def OFF)",default=False,action="store_true")
(options, args) = parser.parse_args()

elist1 = ["Chrom","Pos","Ref","Anc","Alt","Type","Length",'isTv',"isDerived","AnnoType","Consequence","ConsScore","ConsDetail",'GC','CpG','mapAbility20bp','mapAbility35bp','scoreSegDup','priPhCons','mamPhCons','verPhCons','priPhyloP','mamPhyloP','verPhyloP',"GerpN","GerpS",'GerpRS','GerpRSpval','bStatistic','mutIndex','dnaHelT','dnaMGW','dnaProT','dnaRoll','mirSVR-Score','mirSVR-E','mirSVR-Aln','targetScan','fitCons','cHmmTssA','cHmmTssAFlnk','cHmmTxFlnk','cHmmTx','cHmmTxWk','cHmmEnhG','cHmmEnh','cHmmZnfRpts','cHmmHet','cHmmTssBiv','cHmmBivFlnk','cHmmEnhBiv','cHmmReprPC','cHmmReprPCWk','cHmmQuies',"EncExp","EncH3K27Ac","EncH3K4Me1","EncH3K4Me3","EncNucleo","EncOCC","EncOCCombPVal","EncOCDNasePVal","EncOCFairePVal","EncOCpolIIPVal","EncOCctcfPVal","EncOCmycPVal","EncOCDNaseSig","EncOCFaireSig","EncOCpolIISig","EncOCctcfSig","EncOCmycSig","Segway",'tOverlapMotifs','motifDist','motifECount','motifEName','motifEHIPos','motifEScoreChng','TFBS','TFBSPeaks','TFBSPeaksMax',"isKnownVariant","ESP_AF","ESP_AFR","ESP_EUR","TG_AF","TG_ASN","TG_AMR","TG_AFR","TG_EUR",'minDistTSS','minDistTSE',"GeneID","FeatureID","CCDS","GeneName","cDNApos","relcDNApos","CDSpos","relCDSpos","protPos","relProtPos","Domain","Dst2Splice","Dst2SplType","Exon","Intron","oAA","nAA","Grantham","PolyPhenCat","PolyPhenVal","SIFTcat","SIFTval"]
elist2 = ["Chrom","Start","End","GT","QUAL","GQ","DP","AB","FS","QD","Flag","IDClose","Ref","Anc","Alt","Type","Length",'isTv',"isDerived","AnnoType","Consequence","ConsScore","ConsDetail",'GC','CpG','mapAbility20bp','mapAbility35bp','scoreSegDup','priPhCons','mamPhCons','verPhCons','priPhyloP','mamPhyloP','verPhyloP',"GerpN","GerpS",'GerpRS','GerpRSpval','bStatistic','mutIndex','dnaHelT','dnaMGW','dnaProT','dnaRoll','mirSVR-Score','mirSVR-E','mirSVR-Aln','targetScan','fitCons','cHmmTssA','cHmmTssAFlnk','cHmmTxFlnk','cHmmTx','cHmmTxWk','cHmmEnhG','cHmmEnh','cHmmZnfRpts','cHmmHet','cHmmTssBiv','cHmmBivFlnk','cHmmEnhBiv','cHmmReprPC','cHmmReprPCWk','cHmmQuies',"EncExp","EncH3K27Ac","EncH3K4Me1","EncH3K4Me3","EncNucleo","EncOCC","EncOCCombPVal","EncOCDNasePVal","EncOCFairePVal","EncOCpolIIPVal","EncOCctcfPVal","EncOCmycPVal","EncOCDNaseSig","EncOCFaireSig","EncOCpolIISig","EncOCctcfSig","EncOCmycSig","Segway",'tOverlapMotifs','motifDist','motifECount','motifEName','motifEHIPos','motifEScoreChng','TFBS','TFBSPeaks','TFBSPeaksMax',"isKnownVariant","ESP_AF","ESP_AFR","ESP_EUR","TG_AF","TG_ASN","TG_AMR","TG_AFR","TG_EUR",'minDistTSS','minDistTSE',"GeneID","FeatureID","CCDS","GeneName","cDNApos","relcDNApos","CDSpos","relCDSpos","protPos","relProtPos","Domain","Dst2Splice","Dst2SplType","Exon","Intron","oAA","nAA","Grantham","PolyPhenCat","PolyPhenVal","SIFTcat","SIFTval"]
elist = elist1

if not os.path.isdir(options.ancestor_path) or not os.path.exists(options.ancestor_path+"/hsa_emf.index") or not os.path.exists(options.ancestor_path+"/ptr_emf.index"):
  sys.stderr.write("Require valid path to EMF and EMF index files.\n")
  sys.exit()

# SEGWAY SCORING
segway_order = ["D","L0","L1","F0","F1","R0","R1","R2","R3","R4","R5","C0","C1","H3K9me1","TF0","TF1","TF2","TSS","GS","GE0","GE1","GE2","GM0","GM1","E/GM"][::-1]
segway_ranking = dict(zip(segway_order,range(len(segway_order))))

# MAKE AN INDEX OF HUMAN EMF FILES
hsa_index = emf_init_index(options.ancestor_path+"/hsa_emf.index")
# MAKE AN INDEX OF CHIMP EMF FILES
ptr_index = emf_init_index(options.ancestor_path+"/ptr_emf.index")
# CREATE DATASTRUCTURE DESCRIBING CURRENT DATA BLOCK
current = { "data":None, "filename":None, "species":None, 
            "chrom":None, "pos":None,
            "block": None, "bline":None, "bstart":None, "bend":None,
            "bseqs": None, "bpos": None, "bstrands": None, "bgaps": None,
            "bancestors": None, "bref":None, "btype":None }
emf_obj = (current,hsa_index,ptr_index,options.ancestor_path)

if options.refSpecies == "Hsap": CAnc = "SELF-Ptro"
else: CAnc = "SELF-Hsap"

def resetTabixList(filelist):
  newlist = []
  for (filestream,tabix_iter,tabix_next,tabix_chrom,name) in filelist:
    newlist.append((filestream,tabix_iter,tabix_next,None,name))
  return newlist

def mean(valuelist):
  if len(valuelist) > 0:
    return sum(valuelist)/float(len(valuelist))
  else:
    return None

def maxL(valuelist):
  if len(valuelist) > 0: return max(valuelist)
  else: return None

def minL(valuelist):
  if len(valuelist) > 0: return min(valuelist)
  else: return None

def extractEncodeData(filelist,chrom,start,end=None):
  newlist = []
  values = []
  if end == None: end = start
  for encodeTabix in filelist:
    helper_out = []
    for pos in range(start,end+1):
      encodeTabix,encode_out = get_from_tabix(encodeTabix,chrom,pos,rangescore=True,continuous=options.continuous)
      helper_out.append(maxL(map(lambda x:float(x[-1]),encode_out)))
    encode_out = maxL(helper_out)  
    if encode_out != None:
      values.append(encode_out)
    newlist.append(encodeTabix)
  return newlist,maxL(values)
  
def extractEncodeOpenChromData(filelist,chrom,start,end):
  #0: chrom   chr1    varchar(255)    values  Name of the chromosome
  #1: chromStart      10166   int(10) unsigned        range   Start position in chromosome
  #2: chromEnd        10376   int(10) unsigned        range   End position in chromosome
  #3: name    FAIREOnly_1     varchar(255)    values  Optional. Name given to a region (preferably unique). Use . if no name is assigned.
  #4: score   1000    int(10) unsigned        range   Optional. Indicates how dark the peak will be displayed in the browser (1-1000). If '0', the DCC will assign this based on signal value. Ideally average signalValue per base spread between 100-1000.
  #5: strand  .       char(1)         values  Optional. +/- to denote strand or orientation (whenever applicable). Use '.' if no orientation is assigned.
  #6: thickStart      10166   int(10) unsigned        range   Start of where display should be thick (start codon)
  #7: thickEnd        10376   int(10) unsigned        range   End of where display should be thick (stop codon)
  #8: color   10027008        int(10) unsigned        range   RGB color value for peak
  #9: pValue  0.99    float unsigned  range   Fisher's Combined P-Value (-log 10)
  #10: dnaseSignal     0.0016  float unsigned  range   DNaseI HS Signal
  #11: dnasePvalue     0       float unsigned  range   DNaseI HS P-Value (-log 10)
  #12: faireSignal     0.0158  float unsigned  range   FAIRE Signal
  #13: fairePvalue     1.68    float unsigned  range   FAIRE P-Value (-log 10)
  #14: polIISignal     0.0265  float   range   Pol-II Signal
  #15: polIIPvalue     1.51    float   range   Pol-II P-Value (-log 10)
  #16: ctcfSignal      0.0084  float   range   CTCF Signal
  #17: ctcfPvalue      0       float   range   CTCF P-Value (-log 10)
  #18: cmycSignal      0.0336  float   range   c-Myc Signal
  #19: cmycPvalue      0       float   range   c-Myc P-Value (-log 10)
  #20: ocCode  3       int(10) unsigned        range   Open Chromatin (OC) Code

  maxscores = [[],[],[],[],[],[],[],[],[],[],[],[]] # [(20,locCode),(9,lcomb_pvalue),(11,ldnase_pvalue),(13,lfaire_pvalue),(15,lpolII_pvalue),(17,lctcf_pvalue),(19,lcmyc_pvalue),(10,ldnase_signal),(12,lfaire_signal),(14,lpolII_signal),(16,lctcf_signal),(18,lcmyc_signal)]
  newlist = []
  for encodeTabix in filelist:
    encodeTabix,encode_out = get_range_from_tabix(encodeTabix,chrom,start,end,rangescore=True,continuous=options.continuous)
    newlist.append(encodeTabix)
    if len(encode_out) > 0: 
      #print encode_out
      for pos,ind in enumerate([20,9,11,13,15,17,19,10,12,14,16,18]):
        if ind != 20:
          value = maxL(map(lambda x:float(x[ind]),encode_out))
        else:
          value = minL(map(lambda x:float(x[ind]),encode_out))
        if value != None: maxscores[pos].append(value)
  return newlist,map(maxL,maxscores)

def read_grantham(filename):
  grantham = {}
  if os.path.exists(filename):
    infile = open(filename)
    for line in infile:
      fields = line.split()
      if len(fields) == 2:
        AAs = tuple(fields[0].upper().split("-"))
        grantham[AAs]=fields[1]
      else:
        sys.stderr.write("Grantham scores, unexpected line, skipping: %s\n"%line.strip())
    infile.close()
  else:
    sys.stderr.write("Grantham scores input file does not exist: %s\n"%filename)
  return grantham

def read_transcripts(filename):
  anno,names = {},{}
  #global gcount
  recheck = set()
  gene2transcripts = {}
  transcript2geneid = {}
  if os.path.exists(filename):
    infile = gzip.open(filename)
    for line in infile:
      if line.startswith("#"): continue
      fields = line.split()
      if (fields[2] not in ["CDS","exon"]): continue #(fields[1] in ["nonsense_mediated_decay","retained_intron","antisense"]) or 
      cid,gid = None,None
      for ind,elem in enumerate(fields[8:]):
        if elem == "gene_id":
          gid = fields[ind+1+8][1:-2]
        if elem == "transcript_id":
          cid = fields[ind+1+8][1:-2]
          if fields[1].endswith('decay'): 
            recheck.add(cid)
          if gid not in gene2transcripts:
            gene2transcripts[gid] = set()
          gene2transcripts[gid].add(cid)
          transcript2geneid[cid]=gid
        if cid != None and elem == "gene_name":
          names[cid] = fields[ind+1+8][1:-2]
          break
      if cid != None:
        if cid not in anno:
          anno[cid] = [0,0]
        if fields[2] == "CDS":
          anno[cid][1]+=abs(int(fields[3])-int(fields[4]))+1
        elif fields[2] == "exon":
          anno[cid][0]+=abs(int(fields[3])-int(fields[4]))+1
        #if gcount < 10:
          #sys.stderr.write("%s %s\n"%(cid,annotation[cid]))
          #gcount += 1
    infile.close()

  for transID in recheck:
    fixed = False
    exonlength,cdslength = anno[transID]
    for othertranscript in gene2transcripts[transcript2geneid[transID]]:
      if othertranscript != transID:
        oexonlength,ocdslength = anno[othertranscript]
        if exonlength == oexonlength and ocdslength > cdslength:
          anno[transID][1] = ocdslength
          fixed = True
    if options.verbose:
      if not fixed:
        sys.stderr.write('Did not adjust coding length of decay transcript %s (%s): (%d,%d)\n'%(transID,transcript2geneid[transID],exonlength,cdslength))
      else:
        sys.stderr.write('Adjusted coding length of decay transcript %s (%s): (%d,%d) to (%d,%d)\n'%(transID,transcript2geneid[transID],exonlength,cdslength,anno[transID][0],anno[transID][1]))
  return anno,names

def ParseDNAshape(tabixResO,ref,alt):
  tabixRes = list(tabixResO)
  res = []
  if len(ref) == len(alt): #SIMPLE CASE: SNV/COMPLEX SUBSTITUTION
    while len(alt) > 0:
      calt,alt = alt[0],alt[1:]
      cref,ref = ref[0],ref[1:]
      refvalues = None
      altvalues = None
      for i in range(4):
        try: fields = tabixRes.pop(0)
        except: break
        if fields[3] == cref: #ref line
          refvalues = map(lambda x: None if x == "NA" else float(x),fields[4:])
        elif fields[3] == calt: #alt line
          altvalues = map(lambda x: None if x == "NA" else float(x),fields[4:])
      if refvalues != None and altvalues != None:
        res.append(map(lambda (x,y): "NA" if ((x == None) or (y == None)) else "%.2f"%(x-y),zip(altvalues,refvalues)))
  else: #DELETION OR INSERTION
    while len(ref) > 0:
      cref,ref = ref[0],ref[1:]
      refvalues = None
      altvaluesList = []
      for i in range(4):
        try: fields = tabixRes.pop(0)
        except: break
        if fields[3] == cref: #ref line
          refvalues = map(lambda x: None if x == "NA" else float(x),fields[4:])
        else:
          altvaluesList.append(map(lambda x: None if x == "NA" else float(x),fields[4:]))
      if refvalues != None and len(altvaluesList) == 3:
        tmpRes = []
        for hres in map(lambda altvalues: map(lambda (x,y):None if ((x == None) or (y == None)) else (x-y),zip(altvalues,refvalues)), altvaluesList):
          tmpRes.append((sum(map(lambda x: 0 if x == None else abs(x),hres)),hres))
        tmpRes.sort()
        res.append(map(lambda x: "NA" if x == None else "%.2f"%x,tmpRes[-1][1]))
  return res

###################
## ENSEMBL VEP (--> will be replaced by different versions below)
###################
fVRef,fVAlt,fVQUAL,fVGQ,fVDP,fVAB,fVFS,fVQD,fVFlag,fIDClose,fVAlleles,fVESP,fVTG,fVEnd,fVName,fVCoord = 16*[None]
fVChrom = 0   #Uploaded_variation - MODIFIED FOR INDEXING!!!
fVStart = 1   #Location - MODIFIED FOR INDEXING!!!
fVallele = 2  #Allele
fVgene = 3    #Gene
fVfeature = 4 #Feature
fVfeattype= 5 #Feature_type
fVconseq = 6  #Consequence
fVcDNA = 7    #cDNA_position
fVCDSPos = 8  #CDS_position
fVpPOS = 9    #Protein_position 
fVAA = 10     #Amino_acids
fVCodon = 11  #Codons
fVVar = 12    #Existing_variation
fVExtra = 13  #Extra
## Extra column keys:
## CANONICAL    : Indicates if transcript is canonical for this gene
## CCDS         : Indicates if transcript is a CCDS transcript
## HGNC         : HGNC gene identifier
## ENSP         : Ensembl protein identifer
## HGVSc        : HGVS coding sequence name
## HGVSp        : HGVS protein sequence name
## SIFT         : SIFT prediction
## PolyPhen     : PolyPhen prediction
## EXON         : Exon number
## INTRON       : Intron number
## DOMAINS      : The source and identifer of any overlapping protein domains
## MOTIF_NAME   : The source and identifier of a transcription factor binding profile (TFBP) aligned at this position
## MOTIF_POS    : The relative position of the variation in the aligned TFBP
## HIGH_INF_POS : A flag indicating if the variant falls in a high information position of the TFBP
## MOTIF_SCORE_CHANGE : The difference in motif score of the reference and variant sequences for the TFBP
## CELL_TYPE    : List of cell types and classifications for regulatory feature
## IND          : Individual name
## SV           : IDs of overlapping structural variants
## FREQS        : Frequencies of overlapping variants used in filtering

transversions = set([('A','C'),('C','A'),('T','A'),('A','T'),('C','G'),('G','C'),('G','T'),('T','G')])
transitions = set([('C','T'),('T','C'),('G','A'),('A','G')])

def convert_annotation(vepline,Ref,Anc,scores,g1000lines,ESPlines,encode):
  pPCscores,mPCscores,vPCscores,pPPscores,mPPscores,vPPscores,GERPscores,GERPelements,mapscores,mapscores35bp,bscores,segtypes,tfbs,motifs,segDups,GC,CpG,mirSVRvals,targetScan,mutIndex,fitCons,chromHMM,DNAstruct = scores
  res = {}
  res["Chrom"]=vepline[fVChrom]
  res["Ref"]=Ref.upper()
  res["Alt"]=vepline[fVallele].upper()
  
  if len(res["Alt"]) == len(res["Ref"]):
    res["Type"]="SNV"
  elif len(res["Alt"]) > len(res["Ref"]):
    res["Type"]="INS"
  else:
    res["Type"]="DEL"
  
  res["Length"]=str(max(len(res["Ref"]),len(res["Alt"]))-1)
  
  if fVGQ != None: 
    res['QUAL'] = vepline[fVQUAL]
    res['GQ'] = vepline[fVGQ]
    res['DP'] = vepline[fVDP]
    res['GT'] = vepline[fVAlleles]
    res['AB'] = vepline[fVAB]
    res['FS'] = vepline[fVFS]
    res['QD'] = vepline[fVQD]
    res['IDClose'] = vepline[fIDClose]
    res['Flag'] = vepline[fVFlag]
    res["Start"]=vepline[fVStart]
    res["End"]=vepline[fVEnd]
  else:
    res["Pos"]=vepline[fVStart]

  ancestor = Anc.upper()
  for base in ancestor:
    if base not in ['A','C','G','T']:
      ancestor = ""
      break

  if ancestor != "":
    if res["Alt"].upper() != ancestor: res['isDerived'] = "TRUE"
    else: res['isDerived'] = "FALSE"
    res['Anc'] = ancestor

  if res["Type"] == "SNV":
    abase = ancestor if (ancestor != "" and res["Alt"] != ancestor) else res["Ref"]
    if (abase,res["Alt"]) in transitions:
      res['isTv'] = "FALSE"
    elif (abase,res["Alt"]) in transversions:
      res['isTv'] = "TRUE"

  if GC != None: res["GC"]="%.2f"%GC
  if CpG != None: res["CpG"]="%.2f"%CpG

  if vepline[fVgene] != '-': res['GeneID'] = vepline[fVgene]
  if vepline[fVfeature] != '-': res['FeatureID'] = vepline[fVfeature]
  if vepline[fVfeature] in genenames: res['GeneName'] = genenames[vepline[fVfeature]]

  if len(encode) == 6:
    if encode[0] != None: res["EncExp"] = "%.2f"%encode[0]
    if encode[1] != None: res["EncH3K27Ac"] = "%.2f"%encode[1]
    if encode[2] != None: res["EncH3K4Me1"] = "%.2f"%encode[2]
    if encode[3] != None: res["EncH3K4Me3"] = "%.2f"%encode[3]
    if encode[4] != None: res["EncNucleo"] = "%.2f"%encode[4]
    ocCode,comb_pvalue,dnase_pvalue,faire_pvalue,polII_pvalue,ctcf_pvalue,cmyc_pvalue,dnase_signal,faire_signal,polII_signal,ctcf_signal,cmyc_signal = encode[5]
    if ocCode != None: res["EncOCC"] = "%d"%ocCode
    if comb_pvalue != None: res["EncOCCombPVal"] = "%.3f"%comb_pvalue
    if dnase_pvalue != None: res["EncOCDNasePVal"] = "%.3f"%dnase_pvalue
    if faire_pvalue != None: res["EncOCFairePVal"] = "%.3f"%faire_pvalue
    if polII_pvalue != None: res["EncOCpolIIPVal"] = "%.3f"%polII_pvalue
    if ctcf_pvalue != None: res["EncOCctcfPVal"] = "%.3f"%ctcf_pvalue
    if cmyc_pvalue != None: res["EncOCmycPVal"] = "%.3f"%cmyc_pvalue
    if dnase_signal != None: res["EncOCDNaseSig"] = "%.2f"%dnase_signal
    if faire_signal != None: res["EncOCFaireSig"] = "%.2f"%faire_signal
    if polII_signal != None: res["EncOCpolIISig"] = "%.2f"%polII_signal
    if ctcf_signal != None: res["EncOCctcfSig"] = "%.2f"%ctcf_signal
    if cmyc_signal != None: res["EncOCmycSig"] = "%.2f"%cmyc_signal

  if len(segtypes) == 1: 
    res['Segway'] = segtypes[0][-1]  
  elif len(segtypes) > 1:
    helper = []
    for elem in segtypes:
      helper.append((segway_ranking[elem[-1]],elem[-1]))
    helper.sort()
    res['Segway'] = helper[0][-1]

  if res["Type"] == "DEL":
    iPositions = [int(vepline[fVStart])+1]
  else:
    iPositions = [int(vepline[fVStart])]
  if fVEnd != None:
    iPositions.append(int(vepline[fVEnd]))
  
  minscore = None
  maxvalSPLICE = 20
  for iPos in range(iPositions[0],iPositions[-1]+1):
    if vepline[fVfeature].startswith('ENST'):
      try:
        for elem in exonTabix[0].fetch(reference=vepline[fVChrom], start=max(0,iPos-maxvalSPLICE-1), end=iPos+maxvalSPLICE+1):
          elem = elem.rstrip("\n").split('\t')
          if vepline[fVfeature] in elem[8]:
            start,end = int(elem[3]),int(elem[4])
            strand = elem[6] == "+"
            if iPos < start:
              val = (start - iPos,-1,True if strand else False)
            elif iPos >= start and iPos <= end:
              val =  min((iPos - start + 1,1,True if strand else False),(end - iPos + 1,1,False if strand else True))
            else:
              val = (iPos - end,-1,False if strand else True)
            #sys.stderr.write('Transcript ID (%s) matched; val %s: %s\n'%(vepline[fVfeature],val,elem[:100]))
            if (val[0] <= maxvalSPLICE) and ((minscore == None) or (val[0] < minscore[0])): 
              minscore = val
              if minscore[0] == 1: break
          #else:
            #sys.stderr.write('Transcript ID (%s) does not match: %s\n'%(vepline[fVfeature],elem[8][:100]))
      except: pass

  if minscore != None: 
    res["Dst2Splice"]=minscore[0]*minscore[1]
    res["Dst2SplType"]= "ACCEPTOR" if minscore[2] else "DONOR"


  minDistTSS,minDistTSE = None,None
  maxvalDIST = 10000000
  for iPos in set([min(iPositions),max(iPositions)]):
    for cmaxval in (maxvalDIST/10000,maxvalDIST/1000,maxvalDIST/100,maxvalDIST/10,maxvalDIST):
      try:
        for elem in transcriptTabix[0].fetch(reference=vepline[fVChrom], start=max(0,iPos-cmaxval-1), end=iPos+cmaxval+1):
          #sys.stderr.write("%s\n"%(elem))
          elem = elem.rstrip("\n").split('\t')
          start,end = int(elem[1]),int(elem[2])
          if elem[3] == "-": start,end = end,start
          val = min(max(start-iPos+1,iPos-start),maxvalDIST)
          if minDistTSS == None or val < minDistTSS:
            minDistTSS = val

          val = min(max(end-iPos+1,iPos-end),maxvalDIST)
          if minDistTSE == None or val < minDistTSE:
            minDistTSE = val
      except: pass
      if minDistTSS != None and minDistTSE != None: break

  if minDistTSS != None:
    res["minDistTSS"] = str(minDistTSS)
  else:
    res["minDistTSS"] = str(maxvalDIST)

  if minDistTSE != None:
    res["minDistTSE"] = str(minDistTSE)
  else:
    res["minDistTSE"] = str(maxvalDIST)

  res['isKnownVariant']='FALSE'
  
  AFESP,AFR_AF,EUR_AF = eval_ESP_frequencies(ESPlines,res["Ref"],res["Alt"])
  if AFESP != None: res["ESP_AF"]="%.3f"%AFESP
  if AFR_AF != None: res["ESP_AFR"]="%.3f"%AFR_AF
  if EUR_AF != None: res["ESP_EUR"]="%.3f"%EUR_AF

  AF1000g,ASN_AF,AMR_AF,AFR_AF,EUR_AF,isLowCov = eval_dbSNP_frequencies(g1000lines,res["Ref"],res["Alt"])
  if AF1000g != None: res["TG_AF"]="%.3f"%AF1000g
  if ASN_AF != None: res["TG_ASN"]="%.3f"%ASN_AF
  if AMR_AF != None: res["TG_AMR"]="%.3f"%AMR_AF
  if AFR_AF != None: res["TG_AFR"]="%.3f"%AFR_AF
  if EUR_AF != None: res["TG_EUR"]="%.3f"%EUR_AF
  if isLowCov == False and options.filterLowCov: return ""

  if AF1000g != None or AFESP != None: 
    res['isKnownVariant']='TRUE'

  # LOWEST MAPABILITY
  if len(mapscores) == 1: res['mapAbility20bp'] = mapscores[0][-1]
  elif len(mapscores) > 1: 
    helper = map(lambda x:(float(x[-1]),x[-1]),mapscores)
    helper.sort()
    res['mapAbility20bp'] = helper[0][-1]
  if len(mapscores35bp) == 1: res['mapAbility35bp'] = mapscores35bp[0][-1]
  elif len(mapscores35bp) > 1: 
    helper = map(lambda x:(float(x[-1]),x[-1]),mapscores35bp)
    helper.sort()
    res['mapAbility35bp'] = helper[0][-1]

  # HIGHEST PHASTCONS AND PHYLOP SCORES
  if len(pPCscores) == 1: res['priPhCons'] = pPCscores[0][-1]
  elif len(pPCscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),pPCscores)
    helper.sort()
    res['priPhCons'] = helper[-1][-1]
  if len(mPCscores) == 1: res['mamPhCons'] = mPCscores[0][-1]
  elif len(mPCscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),mPCscores)
    helper.sort()
    res['mamPhCons'] = helper[-1][-1]
  if len(vPCscores) == 1: res['verPhCons'] = vPCscores[0][-1]
  elif len(vPCscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),vPCscores)
    helper.sort()
    res['verPhCons'] = helper[-1][-1]
  if len(pPPscores) == 1: res['priPhyloP'] = pPPscores[0][-1]
  elif len(pPPscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),pPPscores)
    helper.sort()
    res['priPhyloP'] = helper[-1][-1]
  if len(mPPscores) == 1: res['mamPhyloP'] = mPPscores[0][-1]
  elif len(mPPscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),mPPscores)
    helper.sort()
    res['mamPhyloP'] = helper[-1][-1]
  if len(vPPscores) == 1: res['verPhyloP'] = vPPscores[0][-1]
  elif len(vPPscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),vPPscores)
    helper.sort()
    res['verPhyloP'] = helper[-1][-1]

  if len(fitCons) == 1: res['fitCons'] = fitCons[0][-1]
  elif len(fitCons) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),fitCons)
    helper.sort()
    res['fitCons'] = helper[0][-1]

  if len(mutIndex) == 1: res['mutIndex'] = mutIndex[0][-1]
  elif len(mutIndex) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),filter(lambda x:x[-1]!="NA",mutIndex))
    helper.sort()
    res['mutIndex'] = helper[-1][-1]

  if len(targetScan) == 1: res['targetScan'] = targetScan[0][-1]
  elif len(targetScan) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),targetScan)
    helper.sort()
    res['targetScan'] = helper[-1][-1][-1]

  if len(mirSVRvals) == 1: 
    res['mirSVR-Score'] = mirSVRvals[0][-1]
    res['mirSVR-E'] = mirSVRvals[0][-2]
    res['mirSVR-Aln'] = mirSVRvals[0][-3]
  elif len(mirSVRvals) > 1:
    helper = map(lambda x:(float(x[-1]),x[-3:]),mirSVRvals)
    helper.sort()
    #sys.stderr.write(str(helper)+"\n")
    res['mirSVR-Score'] = helper[0][-1][-1]
    res['mirSVR-E'] = helper[0][-1][-2]
    res['mirSVR-Aln'] = helper[0][-1][-3]

  if len(chromHMM) == 1: 
    res['cHmmTssA'] = chromHMM[0][-15]
    res['cHmmTssAFlnk'] = chromHMM[0][-14]
    res['cHmmTxFlnk'] = chromHMM[0][-13]
    res['cHmmTx'] = chromHMM[0][-12]
    res['cHmmTxWk'] = chromHMM[0][-11]
    res['cHmmEnhG'] = chromHMM[0][-10]
    res['cHmmEnh'] = chromHMM[0][-9]
    res['cHmmZnfRpts'] = chromHMM[0][-8]
    res['cHmmHet'] = chromHMM[0][-7]
    res['cHmmTssBiv'] = chromHMM[0][-6]
    res['cHmmBivFlnk'] = chromHMM[0][-5]
    res['cHmmEnhBiv'] = chromHMM[0][-4]
    res['cHmmReprPC'] = chromHMM[0][-3]
    res['cHmmReprPCWk'] = chromHMM[0][-2]
    res['cHmmQuies'] = chromHMM[0][-1]
  elif len(chromHMM) > 1:
    helper = map(lambda x:(float(x[-1]),x),chromHMM)
    helper.sort()
    res['cHmmTssA'] = helper[0][-1][-15]
    res['cHmmTssAFlnk'] = helper[0][-1][-14]
    res['cHmmTxFlnk'] = helper[0][-1][-13]
    res['cHmmTx'] = helper[0][-1][-12]
    res['cHmmTxWk'] = helper[0][-1][-11]
    res['cHmmEnhG'] = helper[0][-1][-10]
    res['cHmmEnh'] = helper[0][-1][-9]
    res['cHmmZnfRpts'] = helper[0][-1][-8]
    res['cHmmHet'] = helper[0][-1][-7]
    res['cHmmTssBiv'] = helper[0][-1][-6]
    res['cHmmBivFlnk'] = helper[0][-1][-5]
    res['cHmmEnhBiv'] = helper[0][-1][-4]
    res['cHmmReprPC'] = helper[0][-1][-3]
    res['cHmmReprPCWk'] = helper[0][-1][-2]
    res['cHmmQuies'] = helper[0][-1][-1]
    
  if len(DNAstruct) == 1: 
    res['dnaHelT'] = DNAstruct[0][-4]
    res['dnaMGW'] = DNAstruct[0][-3]
    res['dnaProT'] = DNAstruct[0][-2]
    res['dnaRoll'] = DNAstruct[0][-1]
  elif len(DNAstruct) > 1:
    helper = map(lambda x:(float(x[-1]),x),DNAstruct)
    helper.sort()
    res['dnaHelT'] = helper[0][-1][-4]
    res['dnaMGW'] = helper[0][-1][-3]
    res['dnaProT'] = helper[0][-1][-2]
    res['dnaRoll'] = helper[0][-1][-1]
    
  # MAX GERP N AND GERP S scores
  if len(GERPscores) == 1:
    res['GerpN'] = GERPscores[0][-2]
    res['GerpS'] = GERPscores[0][-1]
  elif len(GERPscores) > 1:
    helper = map(lambda x:(float(x[-2]),x[-2]),GERPscores)
    helper.sort()
    res['GerpN'] = helper[-1][-1]
    helper = map(lambda x:(float(x[-1]),x[-1]),GERPscores)
    helper.sort()
    res['GerpS'] = helper[-1][-1]
    
  # MAX GERP REGION SCORE AND CORRESPONDING P-VALUE
  if len(GERPelements) == 1:
    res['GerpRS'] = GERPelements[0][-2]
    res['GerpRSpval'] = GERPelements[0][-1]
  elif len(GERPelements) > 1:
    helper = map(lambda x:(float(x[-2]),x[-1],x[-2]),GERPelements)
    helper.sort()
    res['GerpRS'] = helper[0][-2]
    res['GerpRSpval'] = helper[-1][-1]

  # MIN B value
  if len(bscores) == 1: res['bStatistic'] = bscores[0][-1]
  elif len(bscores) > 1:
    helper = map(lambda x:(float(x[-1]),x[-1]),bscores)
    helper.sort()
    res['bStatistic'] = helper[-1][0]

  consequences = set(vepline[fVconseq].split(","))
  res["AnnoType"] = vepline[fVfeattype]
  helper = []
  for i in vepline[fVconseq].split(','):
    i = i.replace("_region_variant","")
    i = i.replace("_variant","")
    i = i.replace("_gene","")
    i = i.replace("_transcript","")
    helper.append(i)
  res["ConsDetail"] = ",".join(helper)

  if vepline[fVfeattype] == "Transcript" and (("mature_miRNA_variant" in consequences) or ("non_coding_exon_variant" in consequences)):
    res["AnnoType"] = "NonCodingTranscript"
  elif vepline[fVfeattype] == "Transcript" and (("missense_variant" in consequences) or ("synonymous_variant" in consequences) or ("stop_gained" in consequences) or ("stop_lost" in consequences) or ("initiator_codon_variant" in consequences) or ("stop_retained_variant" in consequences) or ("frameshift_variant" in consequences) or ("inframe_insertion" in consequences) or ("inframe_deletion" in consequences) or ("incomplete_terminal_codon_variant" in consequences)):
    res["AnnoType"] = "CodingTranscript"


  if "stop_gained" in consequences:
    res["Consequence"] = "STOP_GAINED"
    res['ConsScore'] = '8'
  elif ("missense_variant" in consequences) or ("initiator_codon_variant" in consequences):
    res["Consequence"] = "NON_SYNONYMOUS"
    res['ConsScore'] = '7'
  elif ("inframe_insertion" in consequences) or ("inframe_deletion" in consequences):
    res["Consequence"] = "INFRAME"
    res['ConsScore'] = '7'
  elif ("frameshift_variant" in consequences):
    res["Consequence"] = "FRAME_SHIFT"
    res['ConsScore'] = '7'
  elif ("stop_lost" in consequences) or "incomplete_terminal_codon_variant" in consequences:
    res["Consequence"] = "STOP_LOST"
    res['ConsScore'] = '7'
  elif ("splice_donor_variant" in consequences) or ("splice_acceptor_variant" in consequences):
    res["Consequence"] = "CANONICAL_SPLICE"
    res['ConsScore'] = '6'
  elif ("splice_region_variant" in consequences):
    res["Consequence"] = "SPLICE_SITE"
    res['ConsScore'] = '5'
  elif ("non_coding_exon_variant" in consequences) or ("mature_miRNA_variant" in consequences):
    res["Consequence"] = "NONCODING_CHANGE"
    res['ConsScore'] = '5'
  elif ("synonymous_variant" in consequences) or ("stop_retained_variant" in consequences):
    res["Consequence"] = "SYNONYMOUS"
    res['ConsScore'] = '5'
  elif (("feature_truncation" in consequences) or ("feature_elongation" in consequences)):
    if ("coding_sequence_variant" in consequences):
      res["Consequence"] = "UNKNOWN"
      res['ConsScore'] = '5'
    elif ("regulatory_region_variant" in consequences) or ("TF_binding_site_variant" in consequences):
      res["Consequence"] = "REGULATORY"
      res['ConsScore'] = '4'
    elif "5_prime_UTR_variant" in consequences:
      res["Consequence"] = "5PRIME_UTR"
      res['ConsScore'] = '3'
    elif "3_prime_UTR_variant" in consequences:
      res["Consequence"] = "3PRIME_UTR"
      res['ConsScore'] = '2'
    elif "intron_variant" in consequences:
      res["Consequence"] = "INTRONIC"
      res['ConsScore'] = '2'
    elif ("nc_transcript_variant" in consequences):
      res["Consequence"] = "NONCODING_CHANGE"
      res['ConsScore'] = '5'
    else:
      res["Consequence"] = vepline[fVconseq].split('_')[0].upper()
      if res["Consequence"] in ["UPSTREAM","DOWNSTREAM"]: res['ConsScore'] = '1'
      else:
        res["Consequence"] = "UNKNOWN"
        res['ConsScore'] = '0'
      helper = vepline[fVcDNA].split('/')
      if len(helper) == 2:
        vepline[fVcDNA] = '-/'+helper[1]
      else:
        vepline[fVcDNA] = '-'
  elif ("regulatory_region_variant" in consequences) or ("TF_binding_site_variant" in consequences):
    res["Consequence"] = "REGULATORY"
    res['ConsScore'] = '4'
  elif "5_prime_UTR_variant" in consequences:
    res["Consequence"] = "5PRIME_UTR"
    res['ConsScore'] = '3'
  elif "3_prime_UTR_variant" in consequences:
    res["Consequence"] = "3PRIME_UTR"
    res['ConsScore'] = '2'
  elif "intron_variant" in consequences:
    res["Consequence"] = "INTRONIC"
    res['ConsScore'] = '2'
  elif ("coding_sequence_variant" in consequences):
    res["Consequence"] = "UNKNOWN"
    res['ConsScore'] = '5'      
  elif ("nc_transcript_variant" in consequences):
    res["Consequence"] = "NONCODING_CHANGE"
    res['ConsScore'] = '5'
  elif len(consequences) == 1:
      res["Consequence"] = vepline[fVconseq].split('_')[0].upper()
      if res["Consequence"] in ["UPSTREAM","DOWNSTREAM"]: res['ConsScore'] = '1'
      else: res['ConsScore'] = '0'
  else:
    sys.stderr.write("Need simplification: %s %s)\n"%(consequences,vepline[fVconseq]))
    res["Consequence"] = "UNKNOWN"
    res['ConsScore'] = '5'

  if (vepline[fVfeattype] == "-") or (res["Consequence"] in ["UPSTREAM","DOWNSTREAM"]):
    res["AnnoType"] = "Intergenic"

  transcript = vepline[fVfeature]

  if vepline[fVcDNA][0] != "-":
    helper = vepline[fVcDNA].split("/")
    if len(helper) == 2:
      elength = int(helper[-1])
      vepline[fVcDNA] = helper[0]
    else:
      if transcript in annotation:
        elength = annotation[transcript][0]
      else:
        elength = None
    res['cDNApos'] = vepline[fVcDNA].replace('?-','').replace('-?','').split('-')[0]
    if elength != None:
      res['relcDNApos'] = "%.2f"%(min(1.0,float(vepline[fVcDNA].replace('?-','').replace('-?','').split('-')[0])/elength))

  if vepline[fVCDSPos][0] != "-": 
    helper = vepline[fVCDSPos].split("/")
    if len(helper) == 2:
      elength = int(helper[-1])
      vepline[fVCDSPos] = helper[0]
    else:
      if transcript in annotation:
        elength = annotation[transcript][1]
      else:
        elength = None
    res['CDSpos'] = vepline[fVCDSPos].replace('?-','').replace('-?','').split('-')[0]
    if elength != None:
      res['relCDSpos'] = "%.2f"%(min(1.0,float(vepline[fVCDSPos].replace('?-','').replace('-?','').split('-')[0])/elength))

  if vepline[fVpPOS][0] != "-": 
    helper = vepline[fVpPOS].split("/")
    if len(helper) == 2:
      elength = int(helper[-1])
      vepline[fVpPOS] = helper[0]
    else:
      if transcript in annotation:
        elength = annotation[transcript][1]/3.0
      else:
        elength = None
    res['protPos'] = vepline[fVpPOS].replace('?-','').replace('-?','').split('-')[0]
    if elength != None:
      res['relProtPos'] = "%.2f"%(min(1.0,float(vepline[fVpPOS].replace('?-','').replace('-?','').split('-')[0])/elength))

  if vepline[fVAA] != "-": 
    hfields =  vepline[fVAA].split('/')
    if len(hfields) == 2:
      if res["Type"] == "INS":
        if hfields[0] == hfields[1][0]:
          res['oAA'] = "-"
          res['nAA'] = hfields[1][1:]
        elif hfields[0] == hfields[1][-1]:
          res['oAA'] = "-"
          res['nAA'] = hfields[1][:-1]
        else:
          res['oAA'] = hfields[0]
          res['nAA'] = hfields[1]
      elif res["Type"] == "DEL":
        if hfields[1] == hfields[0][0]:
          res['oAA'] = hfields[0][1:]
          res['nAA'] = "-"
        elif hfields[1] == hfields[0][-1]:
          res['oAA'] = hfields[0][:-1]
          res['nAA'] = "-"
        else:
          res['oAA'] = hfields[0]
          res['nAA'] = hfields[1]
      else:
        res['oAA'] = hfields[0]
        res['nAA'] = hfields[1]
      if (hfields[0],hfields[1]) in grantham_matrix:
        res['Grantham'] = str(grantham_matrix[(hfields[0],hfields[1])])
    elif hfields[0] != "X":
      res['oAA'] = hfields[0]
      res['nAA'] = hfields[0]
      #res['Grantham'] = 0

  if len(vepline) > fVExtra:
    for elem in vepline[fVExtra].split(";"):
      hfields = elem.split("=")
      if hfields[0] == "PolyPhen":
        hfields2 = hfields[1].rstrip(")").split("(")
        if hfields2[0] != "unknown":
          res['PolyPhenCat'] = hfields2[0]
          res['PolyPhenVal'] = hfields2[1]
      elif hfields[0] == "SIFT":
        hfields2 = hfields[1].rstrip(")").split("(")
        res['SIFTcat'] = hfields2[0]
        res['SIFTval'] = hfields2[1]
      elif hfields[0] == "CCDS":
        res['CCDS'] = hfields[1]
      elif hfields[0] == "INTRON":
        res['Intron'] = hfields[1]
#        res['tIntron'] = hfields[1].split("/")[1]
#        res['cIntron'] = hfields[1].split("/")[0]
      elif hfields[0] == "EXON":
        res['Exon'] = hfields[1]
#        res['tExon'] = hfields[1].split("/")[1]
#        res['cExon'] = hfields[1].split("/")[0]
      elif hfields[0] == "DOMAINS":
        ncoils,tmhmm,sigp,ndomain,lcompl = False,False,False,False,False
        for dfields in hfields[1].split(','):
          category,name = dfields.split(":")
          if ("_domain" in category) or ("_profile" in category):
            ndomain = True
          else:
            name = name.lower()
            if name == "ncoils": ncoils = True
            elif name == "tmhmm": tmhmm = True
            elif name == "sigp": sigp = True
            elif name == "seg": lcompl = True
        # Implement simple hierarchy of domain annotations:
        if ncoils: res['Domain'] = "ncoils"
        elif tmhmm: res['Domain'] = "tmhmm"
        elif sigp: res['Domain'] = "sigp"
        elif ndomain: res['Domain'] = "ndomain"
        elif lcompl: res['Domain'] = "lcompl"
      elif hfields[0] == "MOTIF_COUNT":
        res['motifECount'] = hfields[1]
#      elif hfields[0] == "MOTIF_POS":
#        res['motifEPos'] = hfields[1]
      elif hfields[0] == "MOTIF_NAME":
        res['motifEName'] = hfields[1]
      elif hfields[0] == "HIGH_INF_POS":
        res['motifEHIPos'] = "TRUE" if "Y" == hfields[1] else "FALSE"
      elif hfields[0] == "MOTIF_SCORE_CHANGE":
        res['motifEScoreChng'] = hfields[1]

  #if 'oAA' in res and (res['nAA'] == "*" or res['oAA'] == "*"):
    #if 'Grantham' not in res: res['Grantham'] = "255"
    #if 'PolyPhenCat' not in res:
      #res['PolyPhenCat'] = "probably_damaging"
      #res['PolyPhenVal'] = "1"
    #if 'SIFTcat' not in res:
      #res['SIFTcat'] = "deleterious"
      #res['SIFTval'] = "0"

  if len(segDups) > 0:
    res['scoreSegDup'] = "%.2f"%(max(map(lambda x:float(x[-1]),segDups)))

  if len(tfbs) > 0:
    for tfbsline in tfbs:
      if ('TFBS' not in res) or (int(res['TFBS']) < int(tfbsline[-6])):
        res['TFBS'] = tfbsline[-6]
        res['TFBSPeaks'] = tfbsline[-4]
        res['TFBSPeaksMax'] = tfbsline[-2]
        #sys.stderr.write('%s\t%s\t%s\t%s\t%s\n'%(res['Chrom'],res['Pos'],res['TFBS'],res['TFBSPeaks'],res['TFBSPeaksMax']))
  if len(motifs) > 0:
    val1,val2 = 0,0
    failed = False
    for motif in motifs:
      sites = motif[-1].split(';')
      res['tOverlapMotifs'] = len(sites)
      ACGT = [0.0,0.0,0.0,0.0]
      for site in sites:
        for ind,value in enumerate(site.split('_')[-4:]):
          ACGT[ind] += float(value[1:])

      if res["Type"] == "INS" or res["Type"] == "DEL":
        val2 += max(ACGT)/len(sites)
      else:
        if res["Ref"] == 'A': val1+=ACGT[0]/len(sites)
        elif res["Ref"] == 'C': val1+=ACGT[1]/len(sites)
        elif res["Ref"] == 'G': val1+=ACGT[2]/len(sites)
        elif res["Ref"] == 'T': val1+=ACGT[3]/len(sites)
        else: failed = True
        
        if res["Alt"] == 'A': val2+=ACGT[0]/len(sites)
        elif res["Alt"] == 'C': val2+=ACGT[1]/len(sites)
        elif res["Alt"] == 'G': val2+=ACGT[2]/len(sites)
        elif res["Alt"] == 'T': val2+=ACGT[3]/len(sites)
        else: failed = True

    if not failed:
      res['motifDist']="%.2f"%(val1-val2)

  return annotation2line(res)

def annotation2line(data,header=False):
  global elist
  if header:
    return "\t".join(elist)+"\n"
  else:
    fstr = ""
    for elem in elist:
      if elem in data: fstr += "%s\t"%(data[elem])
      else: fstr += "NA\t"
  return fstr[:-1]+"\n"

if options.verbose: sys.stderr.write("Parsing ENSEMBL annotation for transcript length...\n")
annotation,genenames = read_transcripts(options.transcripts)

genome_index = read_genome_fastaindex(options.reference+".fai")
#genome_map = get_fasta_mmap(options.reference)
genome = open(options.reference)

if options.verbose: sys.stderr.write("Parsing Grantham scores...\n")
grantham_matrix = read_grantham(options.grantham)

EncodeH3K27Ac = []
if os.path.exists(options.H3K27Ac):
  cpath = "/".join(options.H3K27Ac.split('/')[:-1])
  infile = open(options.H3K27Ac)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeH3K27Ac.append((pysam.Tabixfile(fullname,'r'),None,None,None,"H3K27Ac %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.H3K27Ac)

EncodeH3K4Me1 = []
if os.path.exists(options.H3K4Me1):
  cpath = "/".join(options.H3K4Me1.split('/')[:-1])
  infile = open(options.H3K4Me1)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeH3K4Me1.append((pysam.Tabixfile(fullname,'r'),None,None,None,"H3K4Me1 %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.H3K4Me1)

EncodeH3K4Me3 = []
if os.path.exists(options.H3K4Me3):
  cpath = "/".join(options.H3K4Me3.split('/')[:-1])
  infile = open(options.H3K4Me3)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeH3K4Me3.append((pysam.Tabixfile(fullname,'r'),None,None,None,"H3K4Me3 %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.H3K4Me3)

EncodeExp = []
if os.path.exists(options.expression):
  cpath = "/".join(options.expression.split('/')[:-1])
  infile = open(options.expression)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeExp.append((pysam.Tabixfile(fullname,'r'),None,None,None,"Expression %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.expression)

EncodeOpenChrom = []
if os.path.exists(options.openChrom):
  cpath = "/".join(options.openChrom.split('/')[:-1])
  infile = open(options.openChrom)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeOpenChrom.append((pysam.Tabixfile(fullname,'r'),None,None,None,"openChrom %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.openChrom)

EncodeNucleosome = []
if os.path.exists(options.nucleosome):
  cpath = "/".join(options.nucleosome.split('/')[:-1])
  infile = open(options.nucleosome)
  for filename in infile:
    fullname = cpath+'/'+filename.strip()
    if os.path.exists(fullname):
      EncodeNucleosome.append((pysam.Tabixfile(fullname,'r'),None,None,None,"Nucleosome %s"%fullname))
    else:
      sys.stderr.write("Not found: %s\n"%fullname)
  infile.close()
else:
  sys.stderr.write("Index file not found: %s\n"%options.nucleosome)


G1000Tabix = None
if options.G1000 != "" and (not os.path.exists(options.G1000) or not os.path.exists(options.G1000+".tbi")):
  sys.stderr.write("1000G variants: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.G1000) and os.path.exists(options.G1000+".tbi"):
  if options.verbose: sys.stderr.write('Opening 1000G variants tabix file (%s)...\n'%(options.G1000))
  G1000Tabix = pysam.Tabixfile(options.G1000,'r'),None,None,None,"1000G variants"

ESPTabix = None
if options.ESP != "" and (not os.path.exists(options.ESP) or not os.path.exists(options.ESP+".tbi")):
  sys.stderr.write("ESP annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.ESP) and os.path.exists(options.ESP+".tbi"):
  if options.verbose: sys.stderr.write('ESP annotation variants tabix file (%s)...\n'%(options.ESP))
  ESPTabix = pysam.Tabixfile(options.ESP,'r'),None,None,None,"ESP annotation"


bscoresTabix = None
if options.bscores != "" and (not os.path.exists(options.bscores) or not os.path.exists(options.bscores+".tbi")):
  sys.stderr.write("bscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.bscores) and os.path.exists(options.bscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening B scores tabix file (%s)...\n'%(options.bscores))
  bscoresTabix = pysam.Tabixfile(options.bscores,'r'),None,None,None,"B scores"

pPCscoresTabix = None
if options.primatePhastCONSscores != "" and (not os.path.exists(options.primatePhastCONSscores) or not os.path.exists(options.primatePhastCONSscores+".tbi")):
  sys.stderr.write("primatePhastCONSscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.primatePhastCONSscores) and os.path.exists(options.primatePhastCONSscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening primatePhastCONSscores tabix file (%s)...\n'%(options.primatePhastCONSscores))
  pPCscoresTabix = pysam.Tabixfile(options.primatePhastCONSscores,'r'),None,None,None,"primatePhastCONSscores"

mPCscoresTabix = None
if options.mammalPhastCONSscores != "" and (not os.path.exists(options.mammalPhastCONSscores) or not os.path.exists(options.mammalPhastCONSscores+".tbi")):
  sys.stderr.write("mammalPhastCONSscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mammalPhastCONSscores) and os.path.exists(options.mammalPhastCONSscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening mammalPhastCONSscores tabix file (%s)...\n'%(options.mammalPhastCONSscores))
  mPCscoresTabix = pysam.Tabixfile(options.mammalPhastCONSscores,'r'),None,None,None,"mammalPhastCONSscores"

vPCscoresTabix = None
if options.vertebratePhastCONSscores != "" and (not os.path.exists(options.vertebratePhastCONSscores) or not os.path.exists(options.vertebratePhastCONSscores+".tbi")):
  sys.stderr.write("vertebratePhastCONSscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.vertebratePhastCONSscores) and os.path.exists(options.vertebratePhastCONSscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening vertebratePhastCONSscores tabix file (%s)...\n'%(options.vertebratePhastCONSscores))
  vPCscoresTabix = pysam.Tabixfile(options.vertebratePhastCONSscores,'r'),None,None,None,"vertebratePhastCONSscores"

pPPscoresTabix = None
if options.primatePhyloPscores != "" and (not os.path.exists(options.primatePhyloPscores) or not os.path.exists(options.primatePhyloPscores+".tbi")):
  sys.stderr.write("primatePhyloPscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.primatePhyloPscores) and os.path.exists(options.primatePhyloPscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening primatePhyloPscores tabix file (%s)...\n'%(options.primatePhyloPscores))
  pPPscoresTabix = pysam.Tabixfile(options.primatePhyloPscores,'r'),None,None,None,"primatePhyloPscores"

mPPscoresTabix = None
if options.mammalPhyloPscores != "" and (not os.path.exists(options.mammalPhyloPscores) or not os.path.exists(options.mammalPhyloPscores+".tbi")):
  sys.stderr.write("mammalPhyloPscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mammalPhyloPscores) and os.path.exists(options.mammalPhyloPscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening mammalPhyloPscores tabix file (%s)...\n'%(options.mammalPhyloPscores))
  mPPscoresTabix = pysam.Tabixfile(options.mammalPhyloPscores,'r'),None,None,None,"mammalPhyloPscores"

vPPscoresTabix = None
if options.vertebratePhyloPscores != "" and (not os.path.exists(options.vertebratePhyloPscores) or not os.path.exists(options.vertebratePhyloPscores+".tbi")):
  sys.stderr.write("vertebratePhyloPscores: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.vertebratePhyloPscores) and os.path.exists(options.vertebratePhyloPscores+".tbi"):
  if options.verbose: sys.stderr.write('Opening vertebratePhyloPscores tabix file (%s)...\n'%(options.vertebratePhyloPscores))
  vPPscoresTabix = pysam.Tabixfile(options.vertebratePhyloPscores,'r'),None,None,None,"vertebratePhyloPscores"

mapTabix = None
if options.mapability20bp != "" and (not os.path.exists(options.mapability20bp) or not os.path.exists(options.mapability20bp+".tbi")):
  sys.stderr.write("mapability 20bp: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mapability20bp) and os.path.exists(options.mapability20bp+".tbi"):
  if options.verbose: sys.stderr.write('Opening Mapability 20bp scores tabix file (%s)...\n'%(options.mapability20bp))
  mapTabix = pysam.Tabixfile(options.mapability20bp,'r'),None,None,None,"Mapability 20bp conservation scores"

mapTabix35bp = None
if options.mapability35bp != "" and (not os.path.exists(options.mapability35bp) or not os.path.exists(options.mapability35bp+".tbi")):
  sys.stderr.write("mapability 35bp: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mapability35bp) and os.path.exists(options.mapability35bp+".tbi"):
  if options.verbose: sys.stderr.write('Opening Mapability 35bp scores tabix file (%s)...\n'%(options.mapability35bp))
  mapTabix35bp = pysam.Tabixfile(options.mapability35bp,'r'),None,None,None,"Mapability 35bp conservation scores"

exonTabix = None
if options.exons != "" and (not os.path.exists(options.exons) or not os.path.exists(options.exons+".tbi")):
  sys.stderr.write("Exon annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.exons) and os.path.exists(options.exons+".tbi"):
  if options.verbose: sys.stderr.write('Exon annotation variants tabix file (%s)...\n'%(options.exons))
  exonTabix = pysam.Tabixfile(options.exons,'r'),None,None,None,"Exon annotation"

transcriptTabix = None
if options.exons != "" and (not os.path.exists(options.transcribed) or not os.path.exists(options.transcribed+".tbi")):
  sys.stderr.write("Transcribed range: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.transcribed) and os.path.exists(options.transcribed+".tbi"):
  if options.verbose: sys.stderr.write('Transcribed range tabix file (%s)...\n'%(options.transcribed))
  transcriptTabix = pysam.Tabixfile(options.transcribed,'r'),None,None,None,"Transcribed range annotation"

segwayTabix = None
if options.segway != "" and (not os.path.exists(options.segway) or not os.path.exists(options.segway+".tbi")):
  sys.stderr.write("Segway annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.segway) and os.path.exists(options.segway+".tbi"):
  if options.verbose: sys.stderr.write('Segway annotation variants tabix file (%s)...\n'%(options.segway))
  segwayTabix = pysam.Tabixfile(options.segway,'r'),None,None,None,"Segway annotation"

segDupsTabix = None
if options.segDups != "" and (not os.path.exists(options.segDups) or not os.path.exists(options.segDups+".tbi")):
  sys.stderr.write("segDups annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.segDups) and os.path.exists(options.segDups+".tbi"):
  if options.verbose: sys.stderr.write('segDups annotation variants tabix file (%s)...\n'%(options.segDups))
  segDupsTabix = pysam.Tabixfile(options.segDups,'r'),None,None,None,"segDups annotation"

motifsTabix = None
if options.motifs != "" and (not os.path.exists(options.motifs) or not os.path.exists(options.motifs+".tbi")):
  sys.stderr.write("motifs annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.motifs) and os.path.exists(options.motifs+".tbi"):
  if options.verbose: sys.stderr.write('motifs annotation variants tabix file (%s)...\n'%(options.motifs))
  motifsTabix = pysam.Tabixfile(options.motifs,'r'),None,None,None,"motifs annotation"

GERPelementsTabix = None
if options.GERPelements != "" and (not os.path.exists(options.GERPelements) or not os.path.exists(options.GERPelements+".tbi")):
  sys.stderr.write("GERPelements annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.GERPelements) and os.path.exists(options.GERPelements+".tbi"):
  if options.verbose: sys.stderr.write('GERPelements annotation variants tabix file (%s)...\n'%(options.GERPelements))
  GERPelementsTabix = pysam.Tabixfile(options.GERPelements,'r'),None,None,None,"GERPelements annotation"

GERPscoresTabix = None
if options.GERPscores != "" and (not os.path.exists(options.GERPscores) or not os.path.exists(options.GERPscores+".tbi")):
  sys.stderr.write("GERPscores annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.GERPscores) and os.path.exists(options.GERPscores+".tbi"):
  if options.verbose: sys.stderr.write('GERPscores annotation variants tabix file (%s)...\n'%(options.GERPscores))
  GERPscoresTabix = pysam.Tabixfile(options.GERPscores,'r'),None,None,None,"GERPscores annotation"

tfbsTabix = None
if options.tfbs != "" and (not os.path.exists(options.tfbs) or not os.path.exists(options.tfbs+".tbi")):
  sys.stderr.write("tfbs annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.tfbs) and os.path.exists(options.tfbs+".tbi"):
  if options.verbose: sys.stderr.write('tfbs annotation variants tabix file (%s)...\n'%(options.tfbs))
  tfbsTabix = pysam.Tabixfile(options.tfbs,'r'),None,None,None,"tfbs annotation"

mirSVRTabix = None
if options.mirSVR != "" and (not os.path.exists(options.mirSVR) or not os.path.exists(options.mirSVR+".tbi")):
  sys.stderr.write("mirSVR annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mirSVR) and os.path.exists(options.mirSVR+".tbi"):
  if options.verbose: sys.stderr.write('mirSVR annotation variants tabix file (%s)...\n'%(options.mirSVR))
  mirSVRTabix = pysam.Tabixfile(options.mirSVR,'r'),None,None,None,"mirSVR annotation"

mirTargetScanTabix = None
if options.mirTargetScan != "" and (not os.path.exists(options.mirTargetScan) or not os.path.exists(options.mirTargetScan+".tbi")):
  sys.stderr.write("mirTargetScan annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mirTargetScan) and os.path.exists(options.mirTargetScan+".tbi"):
  if options.verbose: sys.stderr.write('mirTargetScan annotation variants tabix file (%s)...\n'%(options.mirTargetScan))
  mirTargetScanTabix = pysam.Tabixfile(options.mirTargetScan,'r'),None,None,None,"mirTargetScan annotation"

mutabilityIndexTabix = None
if options.mutabilityIndex != "" and (not os.path.exists(options.mutabilityIndex) or not os.path.exists(options.mutabilityIndex+".tbi")):
  sys.stderr.write("mutabilityIndex annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.mutabilityIndex) and os.path.exists(options.mutabilityIndex+".tbi"):
  if options.verbose: sys.stderr.write('mutabilityIndex annotation variants tabix file (%s)...\n'%(options.mutabilityIndex))
  mutabilityIndexTabix = pysam.Tabixfile(options.mutabilityIndex,'r'),None,None,None,"mutabilityIndex annotation"

chromHMM_EpigenomesTabix = None
if options.chromHMM_Epigenomes != "" and (not os.path.exists(options.chromHMM_Epigenomes) or not os.path.exists(options.chromHMM_Epigenomes+".tbi")):
  sys.stderr.write("chromHMM_Epigenomes annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.chromHMM_Epigenomes) and os.path.exists(options.chromHMM_Epigenomes+".tbi"):
  if options.verbose: sys.stderr.write('chromHMM_Epigenomes annotation variants tabix file (%s)...\n'%(options.chromHMM_Epigenomes))
  chromHMM_EpigenomesTabix = pysam.Tabixfile(options.chromHMM_Epigenomes,'r'),None,None,None,"chromHMM_Epigenomes annotation"

DNAstructureTabix = None
if options.DNAstructure != "" and (not os.path.exists(options.DNAstructure) or not os.path.exists(options.DNAstructure+".tbi")):
  sys.stderr.write("DNAstructure annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.DNAstructure) and os.path.exists(options.DNAstructure+".tbi"):
  if options.verbose: sys.stderr.write('DNAstructure annotation variants tabix file (%s)...\n'%(options.DNAstructure))
  DNAstructureTabix = pysam.Tabixfile(options.DNAstructure,'r'),None,None,None,"DNAstructure annotation"

fitConsTabix = None
if options.fitCons != "" and (not os.path.exists(options.fitCons) or not os.path.exists(options.fitCons+".tbi")):
  sys.stderr.write("fitCons annotation: Require valid path to compressed tabix file and tabix index file.\n")
elif os.path.exists(options.fitCons) and os.path.exists(options.fitCons+".tbi"):
  if options.verbose: sys.stderr.write('fitCons annotation variants tabix file (%s)...\n'%(options.fitCons))
  fitConsTabix = pysam.Tabixfile(options.fitCons,'r'),None,None,None,"fitCons annotation"


if os.path.exists(options.vep) or options.pipe:

  if not options.pipe: vepinput = open(options.vep)
  else: vepinput = sys.stdin
  fline = True
  lchrom,lpos,llines = None,None,[]
  lallele = False,None

  pPCscores,mPCscores,vPCscores, pPPscores,mPPscores,vPPscores = None,None,None, None,None,None
  hp, hAnc, mapscores,mapscores35bp,bscores, segtypes,g1000line,encode = None, None, None,None,None, None,None,None
  tfbs,motifs,segDups, GERPscores,GERPelements = None,None,None, None,None
  mirSVR,TargetScan,mutabilityIndex,fitCons,chromHMM_Epigenomes,DNAstructure = None,None,None, None,None,None

  g1000line,ESPline = None,None

  #count = 0
  for vepline in vepinput:
    #count += 1
    #if count > 10: sys.exit()
    if vepline.startswith('##'): continue
    if fline:
      vepfields = vepline.rstrip("\n").split('\t')
      if len(vepfields) == 14:
        fVRef,fVAlt,fVQUAL,fVGQ,fVDP,fVAB,fVFS,fVQD,fVFlag,fVAlleles,fVESP,fVTG,fVEnd,fVName,fVCoord = 15*[None]
        fVChrom = 0   #Uploaded_variation - MODIFIED FOR INDEXING!!!
        fVStart = 1   #Location - MODIFIED FOR INDEXING!!!
        fVallele = 2  #Allele
        fVgene = 3    #Gene
        fVfeature = 4 #Feature
        fVfeattype= 5 #Feature_type
        fVconseq = 6  #Consequence
        fVcDNA = 7    #cDNA_position
        fVCDSPos = 8  #CDS_position
        fVpPOS = 9    #Protein_position 
        fVAA = 10     #Amino_acids
        fVCodon = 11  #Codons
        fVVar = 12    #Existing_variation
        fVExtra = 13  #Extra
        elist = elist1
      elif len(vepfields) == 17:
        fVRef,fVAlt,fVQUAL,fVGQ,fVDP,fVAB,fVFS,fVQD,fVFlag,fVAlleles,fVESP,fVTG = 12*[None]
        fVChrom = 0   #Location
        fVStart = 1   #Location
        fVEnd = 2     #Location
        fVName = 3    #Uploaded_variation
        fVCoord = 4   #Location
        fVallele = 5  #Allele
        fVgene = 6    #Gene
        fVfeature = 7 #Feature
        fVfeattype= 8 #Feature_type
        fVconseq = 9  #Consequence
        fVcDNA = 10   #cDNA_position
        fVCDSPos = 11 #CDS_position
        fVpPOS = 12   #Protein_position 
        fVAA = 13     #Amino_acids
        fVCodon = 14  #Codons
        fVVar = 15    #Existing_variation
        fVExtra = 16  #Extra
        elist = elist1
      elif len(vepfields) == 30:
        fVRef,fVAlt,fVQUAL,fVGQ,fVDP,fVAB,fVFS,fVQD,fVFlag,fIDClose,fVAlleles,fVESP,fVTG = range(13)
        fVChrom = 13   #Location
        fVStart = 14   #Location
        fVEnd = 15     #Location
        fVName = 16    #Uploaded_variation
        fVCoord = 17   #Location
        fVallele = 18  #Allele
        fVgene = 19    #Gene
        fVfeature = 20 #Feature
        fVfeattype= 21 #Feature_type
        fVconseq = 22  #Consequence
        fVcDNA = 23    #cDNA_position
        fVCDSPos = 24  #CDS_position
        fVpPOS = 25    #Protein_position 
        fVAA = 26      #Amino_acids
        fVCodon = 27   #Codons
        fVVar = 28     #Existing_variation
        fVExtra = 29   #Extra
        elist = elist2
      else:
        sys.stderr.write("Unknown input data format.")
        sys.exit()
      fline = False
      if not options.noheader: sys.stdout.write('#'+annotation2line({},True))

    if vepline.startswith("#") or vepline.startswith("Chrom"): continue
    else:
      vepfields = vepline.rstrip("\n").split('\t')
      if len(vepfields) > 3:
        chrom,pos,alleleVal=vepfields[fVChrom],vepfields[fVStart],vepfields[fVallele]
        alleleVal = alleleVal.upper()
        
        pos = int(pos.split('-')[0])
        alleles = []
        if fVName != None:
          alleles = vepfields[fVName].split("_")[-1].split("/")
        start,end = pos,pos
        is_indel = False
        is_insertion = False
        if len(alleles) == 2 and alleles[-1] == alleleVal:
          altcoord = True
          if alleles[0] == "-":
            href = get_DNA(chrom,pos,pos,genome,genome_index)
            alleles[0] = href
            alleles[-1] = href+alleles[-1]
            if fVRef != None: vepfields[fVRef] = alleles[0]
            vepfields[fVallele] = alleles[-1]
            altcoord = False
            start = int(vepfields[fVStart])
            end = int(vepfields[fVEnd])
          elif alleles[1] == "-":
            href = get_DNA(chrom,pos-1,pos-1,genome,genome_index)
            alleles[0] = href+alleles[0]
            alleles[-1] = href
            if fVRef != None: vepfields[fVRef] = alleles[0]
            vepfields[fVallele] = alleles[-1]
            altcoord = False
            start = int(vepfields[fVStart])-1
            pos = start
            vepfields[fVStart] = str(start)
            end = int(vepfields[fVEnd])
          
          if (len(alleles[0]) != len(alleles[1])):
            is_indel = True
            alleleVal = tuple(alleles)
            if len(alleles[1]) > len(alleles[0]): # INSERTION
              if altcoord:
                start = pos
                end = pos+1
              is_insertion = True
            else: # DELETION
              if altcoord:
                start = pos
                end = pos+len(alleles[0])-1
          elif len(alleles[0]) != 1:
            alleleVal = tuple(alleles)
            if altcoord:
              start = pos
              end = pos+len(alleles[0])-1

        vepfields[fVStart]=str(start)
        if fVEnd != None: vepfields[fVEnd]=str(end)
        allele = is_indel,alleleVal

        if lchrom == chrom and lpos == pos and lallele == allele:
          llines.append(vepfields)
        else:
          if len(llines) > 0:
            hmotifline = None
            hmotifscore = None
            motifcount = 0
            veplines = [[],[],[],[]]
            for cline in llines:
              if cline[fVfeattype] == "MotifFeature":
                motifcount += 1
                if hmotifline == None: hmotifline = cline                
                for elem in cline[fVExtra].split(';'): 
                  if elem.startswith("MOTIF_SCORE_CHANGE="): 
                    score = abs(float(elem.split('=')[-1]))
                    if hmotifscore == None or hmotifscore < score:
                      hmotifscore = score

              else:
                if options.random:
                  veplines[0].append(cline)
                else:
                  if ("coding" in cline[fVconseq]) or ("missense" in cline[fVconseq]) or ("synonymous" in cline[fVconseq]) or ("stop" in cline[fVconseq]) or ("mature" in cline[fVconseq]) or ("splice" in cline[fVconseq]) or ("initiator_codon" in cline[fVconseq]) or ("frame" in cline[fVconseq]) or ("terminal_codon" in cline[fVconseq]):
                    veplines[0].append(cline)
                  elif ("utr" in cline[fVconseq]) or ("regulatory" in cline[fVconseq]):
                    veplines[1].append(cline)
                  elif "intronic" in cline[fVconseq] or ("upstream" in cline[fVconseq]) or ("downstream" in cline[fVconseq]):
                    veplines[2].append(cline)
                  else:
                    veplines[3].append(cline)

            if not options.all:
              hvepline = None
              for veps in veplines:
                if len(veps) > 0:
                  hvepline = random.sample(veps,1)[0]
                  if hmotifscore != None: hvepline[fVExtra] = ";".join([hvepline[fVExtra],hmotifline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
                  break
              if hvepline == None:
                hvepline = hmotifline
                hvepline[fVExtra] = ";".join([hvepline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
              sys.stdout.write(convert_annotation(hvepline,hp,hAnc,scores,g1000line,ESPline,encode))
            else:
              hvepline = []
              for veps in veplines:
                for hline in veps:
                  if hmotifscore != None: hline[fVExtra] = ";".join([hline[fVExtra],hmotifline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
                  hvepline.append(hline)
              if len(hvepline) == 0:
                hvepline.append(hmotifline)
                hvepline[0][fVExtra] = ";".join([hvepline[0][fVExtra],"MOTIF_COUNT=%d"%(motifcount)])

              for hline in hvepline:
                sys.stdout.write(convert_annotation(hline,hp,hAnc,scores,g1000line,ESPline,encode))


          if ((chrom != lchrom) or (lpos != pos) or (allele[0] != lallele[0]) or (allele[0] and allele[1] != lallele[1])):
            if fVRef != None: hp = vepfields[fVRef]
            elif len(alleles) != 0: hp = alleles[0]
            else: hp = get_DNA(chrom,pos,pos,genome,genome_index)

            GC,CpG = count_GC_CpG(chrom,start-75,end+75,genome,genome_index)
            hAnc = ""
            if is_insertion: irange = range(start,end)
            else: irange = range(start,end+1)
            for ipos in irange:
              hAnc_ = format_ancestor(emf_get_base_at_position(emf_obj,options.refSpecies,chrom,ipos,CAnc,True))
              if hAnc_ == "N/A":
                hAnc_ = format_ancestor(emf_get_base_at_position(emf_obj,options.refSpecies,chrom,ipos,"SELF-Ggor",True))
                if hAnc_ == "N/A":
                  hAnc_ = format_ancestor(emf_get_base_at_position(emf_obj,options.refSpecies,chrom,ipos,"SELF-Pabe",True))
                  if hAnc_ == "N/A":
                    hAnc_ = format_ancestor(emf_get_base_at_position(emf_obj,options.refSpecies,chrom,ipos,"SELF-Mmul",True))
                    if hAnc_ == "N/A":
                      hAnc_ = format_ancestor(emf_get_base_at_position(emf_obj,options.refSpecies,chrom,ipos,"SELF-Cjac",True))
                    else: hAnc += hAnc_.replace(",","").replace("-","")
                  else: hAnc += hAnc_.replace(",","").replace("-","")
                else: hAnc += hAnc_.replace(",","").replace("-","")
              else: hAnc += hAnc_.replace(",","").replace("-","")
            
            # FOR DELETIONS WE DO NOT WANT TO CONSIDER THE BASE BEFORE THE EVENT:
            if not is_insertion and is_indel: start+=1

            pPCscoresTabix,pPCscores = get_range_from_tabix(pPCscoresTabix,chrom,start,end,continuous=options.continuous)
            mPCscoresTabix,mPCscores = get_range_from_tabix(mPCscoresTabix,chrom,start,end,continuous=options.continuous)
            vPCscoresTabix,vPCscores = get_range_from_tabix(vPCscoresTabix,chrom,start,end,continuous=options.continuous)
            pPPscoresTabix,pPPscores = get_range_from_tabix(pPPscoresTabix,chrom,start,end,continuous=options.continuous)
            mPPscoresTabix,mPPscores = get_range_from_tabix(mPPscoresTabix,chrom,start,end,continuous=options.continuous)
            vPPscoresTabix,vPPscores = get_range_from_tabix(vPPscoresTabix,chrom,start,end,continuous=options.continuous)
            GERPscoresTabix,GERPscores = get_range_from_tabix(GERPscoresTabix,chrom,start,end,continuous=options.continuous)
            GERPelementsTabix,GERPelements = get_range_from_tabix(GERPelementsTabix,chrom,start,end,rangescore=True,continuous=options.continuous)

            mapTabix,mapscores = get_range_from_tabix(mapTabix,chrom,start,end,rangescore=True,continuous=options.continuous)
            mapTabix35bp,mapscores35bp = get_range_from_tabix(mapTabix35bp,chrom,start,end,rangescore=True,continuous=options.continuous)
            bscoresTabix,bscores = get_range_from_tabix(bscoresTabix,chrom,start,end,rangescore=True,continuous=options.continuous)
            segwayTabix,segtypes = get_range_from_tabix(segwayTabix,chrom,start,end,rangescore=True,continuous=options.continuous)

            tfbsTabix,tfbs = get_range_from_tabix(tfbsTabix,chrom,start,end,continuous=options.continuous)
            
            mirSVRTabix,mirSVR = get_range_from_tabix(mirSVRTabix,chrom,start,end,continuous=options.continuous,multirange=True)
            mirTargetScanTabix,TargetScan = get_range_from_tabix(mirTargetScanTabix,chrom,start,end,continuous=options.continuous,multirange=True)
            mutabilityIndexTabix,mutabilityIndex = get_range_from_tabix(mutabilityIndexTabix,chrom,start,end,continuous=options.continuous)
            fitConsTabix,fitCons = get_range_from_tabix(fitConsTabix,chrom,start,end,continuous=options.continuous)
            chromHMM_EpigenomesTabix,chromHMM_Epigenomes = get_range_from_tabix(chromHMM_EpigenomesTabix,chrom,start,end,continuous=options.continuous)
            DNAstructureTabix,DNAstructure = get_range_from_tabix(DNAstructureTabix,chrom,start,end,continuous=options.continuous,multirange=True)
            DNAstructureFinal = ParseDNAshape(DNAstructure,hp,alleleVal)
  
            motifsTabix,motifs = get_range_from_tabix(motifsTabix,chrom,start,end,continuous=options.continuous,multirange=True)
            segDupsTabix,segDups = get_range_from_tabix(segDupsTabix,chrom,start,end,rangescore=True,continuous=options.continuous,multirange=True)

            scores = (pPCscores,mPCscores,vPCscores,pPPscores,mPPscores,vPPscores,GERPscores,GERPelements,mapscores,mapscores35bp,bscores,segtypes,tfbs,motifs,segDups,GC,CpG,mirSVR,TargetScan,mutabilityIndex,fitCons,chromHMM_Epigenomes,DNAstructureFinal)

            G1000Tabix,g1000line = get_from_tabix(G1000Tabix,chrom,pos,continuous=options.continuous)
            ESPTabix,ESPline = get_from_tabix(ESPTabix,chrom,pos,continuous=options.continuous)

            EncodeExp,ExpValue = extractEncodeData(EncodeExp,chrom,start,end)
            EncodeH3K27Ac,H3K27AcValue = extractEncodeData(EncodeH3K27Ac,chrom,start,end)
            EncodeH3K4Me1,H3K4Me1Value = extractEncodeData(EncodeH3K4Me1,chrom,start,end)
            EncodeH3K4Me3,H3K4Me3Value = extractEncodeData(EncodeH3K4Me3,chrom,start,end)
            EncodeNucleosome,NucleosomeValue = extractEncodeData(EncodeNucleosome,chrom,start,end)
            EncodeOpenChrom,OpenChromValues = extractEncodeOpenChromData(EncodeOpenChrom,chrom,start,end)
            encode = (ExpValue,H3K27AcValue,H3K4Me1Value,H3K4Me3Value,NucleosomeValue,OpenChromValues)

          lchrom,lpos,lallele,llines = chrom,pos,(is_indel,alleleVal),[vepfields]
  if len(llines) > 0:
    hmotifline = None
    hmotifscore = None
    motifcount = 0
    veplines = [[],[],[],[]]
    for cline in llines:
      if cline[fVfeattype] == "MotifFeature":
        motifcount += 1
        if hmotifline == None: hmotifline = cline                
        for elem in cline[fVExtra].split(';'): 
          if elem.startswith("MOTIF_SCORE_CHANGE="): 
            score = abs(float(elem.split('=')[-1]))
            if hmotifscore == None or hmotifscore < score:
              hmotifscore = score
              hmotifline = cline
      else:
        if options.random:
          veplines[0].append(cline)
        else:
          if ("coding" in cline[fVconseq]) or ("missense" in cline[fVconseq]) or ("synonymous" in cline[fVconseq]) or ("stop" in cline[fVconseq]) or ("mature" in cline[fVconseq]) or ("splice" in cline[fVconseq]) or ("initiator_codon" in cline[fVconseq]) or ("frame" in cline[fVconseq]) or ("terminal_codon" in cline[fVconseq]):
            veplines[0].append(cline)
          elif ("utr" in cline[fVconseq]) or ("regulatory" in cline[fVconseq]):
            veplines[1].append(cline)
          elif "intronic" in cline[fVconseq] or ("upstream" in cline[fVconseq]) or ("downstream" in cline[fVconseq]):
            veplines[2].append(cline)
          else:
            veplines[3].append(cline)

    if not options.all:
      vepline = None
      for veps in veplines:
        if len(veps) > 0:
          vepline = random.sample(veps,1)[0]
          if hmotifscore != None: vepline[fVExtra] = ";".join([vepline[fVExtra],hmotifline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
          break
      if vepline == None:
        vepline = hmotifline
        vepline[fVExtra] = ";".join([vepline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
      sys.stdout.write(convert_annotation(vepline,hp,hAnc,scores,g1000line,ESPline,encode))
    else:
      vepline = []
      for veps in veplines:
        for hline in veps:
          if hmotifscore != None: hline[fVExtra] = ";".join([hline[fVExtra],hmotifline[fVExtra],"MOTIF_COUNT=%d"%(motifcount)])
          vepline.append(hline)
      if len(vepline) == 0:
        vepline.append(hmotifline)
        vepline[0][fVExtra] = ";".join([vepline[0][fVExtra],"MOTIF_COUNT=%d"%(motifcount)])

      for hline in vepline:
        sys.stdout.write(convert_annotation(hline,hp,hAnc,scores,g1000line,ESPline,encode))

  if not options.pipe: vepinput.close()
