#!/usr/bin/env python
# -*- coding: ASCII -*-

"""

:Author: Martin Kircher
:Contact: mkircher@uw.edu
:Date: *31.08.2014
"""

import sys, os
import random
from optparse import OptionParser
import math

#Cl Name Min Max 99% Factor 99%/Max    
#28	EncExp	0.41	235313	593.36	396.58
#29	EncH3K27Ac	0.04	223899	84.24	2657.87
#30	EncH3K4Me1	0.04	6856.88	40.4	169.72
#31	EncH3K4Me3	0.04	52815	72.88	724.68
#32	EncNucleo	0	19045	3.9	4883.33
#33	EncOCC	1	4	4	-
#34	EncOCCombPVal	0	16	16	-
#35	EncOCDNasePVal	0	16	16	-
#36	EncOCFairePVal	0	5.4	16	-
#37	EncOCpolIIPVal	0	16	15.1	-
#38	EncOCctcfPVal	0	16	16	-
#39	EncOCmycPVal	0	16	5.44	-
#40	EncOCDNaseSig	0	2.61	0.51	5.12
#41	EncOCFaireSig	0	0.64	0.04	16.00
#42	EncOCpolIISig	0	10.08	0.25	40.32
#43	EncOCctcfSig	0	14.96	3.09	4.84
#44	EncOCmycSig	0	1.76	0.17	10.35
#
#("GC","0.418482"),("CpG","0.0244861")
#("priPhCons","0.114644"),("mamPhCons","0.0794826"),("verPhCons","0.0939182")
#("priPhyloP","-0.0331088"),("mamPhyloP","-0.0383832"),("verPhyloP","0.0169746")
#("GerpN","1.90901"),("GerpS","-0.20019"),("GerpRS","0"),("GerpRSpval","1"),("bStatistic","800.261"),
#("PolyPhenCat","unknown"),("PolyPhenVal",0.404105),("SIFTcat","unknown"),("SIFTval",0.218367)]
#
#      TFBS           TFBSPeaks        TFBSPeaksMax     
# Min.   :  1.000   Min.   :  1.000   Min.   :   9.717  
# 1st Qu.:  1.000   1st Qu.:  1.000   1st Qu.:  36.957  
# Median :  2.000   Median :  2.000   Median :  55.573  
# Mean   :  5.008   Mean   :  6.906   Mean   :  97.565  
# 3rd Qu.:  4.000   3rd Qu.:  6.000   3rd Qu.: 102.889  
# Max.   :123.000   Max.   :275.000   Max.   :1458.610
#
# Capped at 99th percentile:
#
#> quantile(data$TFBS,0.99)
#48 
#> quantile(data$TFBSPeaks,0.99)
#81 
#> quantile(data$TFBSPeaksMax,0.99)
#542.0574
#
# lowest p-value other than 0 (GerpRSpval): 4.94066e-324

# Domain: ["ncoils","tmhmm","sigp","ndomain","lcompl"]

transform_sm = [("Chrom",("REMOVE",None,None)),
                ("Pos",("REMOVE",None,None))]
transform_ext = [("Chrom",("REMOVE",None,None)),
                 ("Start",("REMOVE",None,None)),
                 ("End",("REMOVE",None,None)),
                 ("GT",("REMOVE",None,None)),
                 ("QUAL",("REMOVE",None,None)),
                 ("GQ",("REMOVE",None,None)),
                 ("DP",("REMOVE",None,None)),
                 ("AB",("REMOVE",None,None)),
                 ("FS",("REMOVE",None,None)),
                 ("QD",("REMOVE",None,None)),
                 ("Flag",("REMOVE",None,None)),
                 ("IDClose",("REMOVE",None,None))]

transform_p2 = [("Ref",("DROP",None,None)),
            ("Anc",("REMOVE",None,None)),
            ("Alt",("DROP",None,None)),
            ("Type",("DROP",None,None)),
            ("Length",("DROP",None,None)),
            ("isTv",("TV",None,None)),
            ("isDerived",("REMOVE",None,None)),
            ("AnnoType",("REMOVE",None,None)),
            ("Consequence",("SET",None,None)),
            ("ConsScore",("REMOVE",None,None)),
            ("ConsDetail",("REMOVE",None,None)),
            ("GC",(None,"0.42",None)),
            ("CpG",(None,"0.02",None)),
            ("mapAbility20bp",(None,"0",None)),
            ("mapAbility35bp",(None,"0",None)),
            ("scoreSegDup",("SPECIAL",None,None)),
            ("priPhCons",(None,"0.115",None)),
            ("mamPhCons",(None,"0.079",None)),
            ("verPhCons",(None,"0.094",None)),
            ("priPhyloP",(None,"-0.033",None)),
            ("mamPhyloP",(None,"-0.038",None)),
            ("verPhyloP",(None,"0.017",None)),
            ("GerpN",(None,"1.91",None)),
            ("GerpS",(None,"-0.200",None)),
            ("GerpRS",(None,"0",None)),
            ("GerpRSpval",("LOG","1","4.94066e-324")),
            ("bStatistic",(None,"800",None)),
            ("mutIndex",(None,"0",None)), # NEW; impute missing values to 0 [no effect]
            ("dnaHelT",(None,"0",None)), # NEW; impute missing values to 0 [no effect]
            ("dnaMGW",(None,"0",None)), # NEW; impute missing values to 0 [no effect]
            ("dnaProT",(None,"0",None)), # NEW; impute missing values to 0 [no effect]
            ("dnaRoll",(None,"0",None)), # NEW; impute missing values to 0 [no effect]
            ("mirSVRs",(None,"0","IND")), # NEW; add indicator for missing
            ("mirSVRe",(None,"0",None)), # NEW
            ("mirSVRa",(None,"0",None)), # NEW
            ("targetScan",(None,"0","IND")), # NEW; add indicator for missing
            ("fitCons",("REMOVE","0.061782",None)), # NEW; imputation to value used for N bases: 0.061782
            ("cHmmTssA",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmTssAFlnk",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmTxFlnk",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmTx",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmTxWk",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmEnhG",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmEnh",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmZnfRpts",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmHet",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmTssBiv",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmBivFlnk",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmEnhBiv",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmReprPC",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmReprPCWk",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("cHmmQuies",(None,"0.0667",None)), # NEW; imputation to all states having the same likelihood
            ("EncExp",("0",593.36,"593.36")),
            ("EncH3K27Ac",("0",84.24,"84.24")),
            ("EncH3K4Me1",("0",40.4,"40.4")),
            ("EncH3K4Me3",("0",72.88,"72.88")),
            ("EncNucleo",("0",3.9,"3.9")),
            ("EncOCC",(None,"5",None)),
            ("EncOCCombPVal",(None,"0",None)),
            ("EncOCDNasePVal",(None,"0",None)),
            ("EncOCFairePVal",(None,"0",None)),
            ("EncOCpolIIPVal",(None,"0",None)),
            ("EncOCctcfPVal",(None,"0",None)),
            ("EncOCmycPVal",(None,"0",None)),
            ("EncOCDNaseSig",("0",0.51,"0.51")),
            ("EncOCFaireSig",("0",0.04,"0.04")),
            ("EncOCpolIISig",("0",0.25,"0.25")),
            ("EncOCctcfSig",("0",3.09,"3.09")),
            ("EncOCmycSig",("0",0.17,"0.17")),
            ("EncSegway",(None,"unknown",None)),
            ("tOverlapMotifs",(None,"0",None)),
            ("motifDist",(None,"0",None)),
            ("motifECount",(None,"0",None)),
            ("motifEName",("REMOVE",None,None)),
            ("motifEHIPos",(None,"0",None)),
            ("motifEScoreChng",(None,"0",None)),
            ("TFBS",("0",48,"48")),
            ("TFBSPeaks",("0",81,"81")),
            ("TFBSPeaksMax",("0",542.0574,"542.0574")),
            ("isKnownVariant",("REMOVE","0",None)),
            ("ESP_AF",("REMOVE","0",None)),
            ("ESP_AFR",("REMOVE","0",None)),
            ("ESP_EUR",("REMOVE","0",None)),
            ("TG_AF",("REMOVE","0",None)),
            ("TG_ASN",("REMOVE","0",None)),
            ("TG_AMR",("REMOVE","0",None)),
            ("TG_AFR",("REMOVE","0",None)),
            ("TG_EUR",("REMOVE","0",None)),
            ("minDistTSS",("LOG","10000000",None)),
            ("minDistTSE",("LOG","10000000",None)),
            ("GeneID",("REMOVE",None,None)),
            ("FeatureID",("REMOVE",None,None)),
            ("CCDS",("REMOVE",None,None)),
            ("GeneName",("REMOVE",None,None)),
            ("cDNApos",(None,"0","IND")),
            ("relcDNApos",(None,"0",None)),
            ("CDSpos",(None,"0","IND")),
            ("relCDSpos",(None,"0",None)),
            ("protpos",(None,"0","IND")),
            ("relProtpos",(None,"0",None)),
            ("Domain",(None,"UD",None)), #NEW; impute to missing categorical
            ("Dst2Splice",(None,"0",None)),
            ("Dst2SplType",(None,"unknown",None)),
            ("Exon",("REMOVE",None,None)),
            ("Intron",("REMOVE",None,None)),
            ("oAA",(None,"unknown",None)),
            ("nAA",(None,"unknown",None)),
            ("Grantham",(None,"0","IND")),
            ("PolyPhenCat",("NS","unknown",None)),
            ("PolyPhenVal",("NS","0.404","IND")),
            ("SIFTcat",("NS","unknown",None)),
            ("SIFTval",("NS","0.22","IND"))]

categorical = { "Type" : ["SNV","INS","DEL"],
                "Consequence" : ["3PRIME_UTR","5PRIME_UTR","DOWNSTREAM","INTERGENIC","INTRONIC","NONCODING_CHANGE","INFRAME","FRAME_SHIFT","NON_SYNONYMOUS","REGULATORY","CANONICAL_SPLICE","SPLICE_SITE","STOP_GAINED","STOP_LOST","SYNONYMOUS","UNKNOWN","UPSTREAM"], 
                "EncSegway":["C0","C1","D","E/GM","F0","F1","GE0","GE1","GE2","GM0","GM1","GS","H3K9me1","L0","L1","R0","R1","R2","R3","R4","R5","TF0","TF1","TF2","TSS","unknown"],
                #"Anc" : ["A","C","G","T","N"],
                "Ref" : ["A","C","G","T","N"],
                "Alt" : ["A","C","G","T","N"],
                "Domain": ["ncoils","tmhmm","sigp","ndomain","lcompl","UD"],
                "Dst2SplType" : ["DONOR","ACCEPTOR","unknown"],
                "PolyPhenCat" : ["benign","possibly_damaging","probably_damaging","unknown","UD"],
                "SIFTcat" : ["deleterious","tolerated","unknown","UD"],
                "oAA" : ["*","A","C","D","E","F","G","H","I","K","L","M","N","P","Q","R","S","T","V","W","Y","unknown"],
                "nAA" : ["*","A","C","D","E","F","G","H","I","K","L","M","N","P","Q","R","S","T","V","W","Y","unknown"] }

consequence_set = set(categorical["Consequence"])
transitions = set([('C','T'),('T','C'),('G','A'),('A','G')])
transversions = set([('A','C'),('C','A'),('T','A'),('A','T'),('C','G'),('G','C'),('G','T'),('T','G')])
validbases = set(["A","C","G","T","N"])
validaminoacids = set(["*","A","C","D","E","F","G","H","I","K","L","M","N","P","Q","R","S","T","V","W","Y","unknown"])

prodlst = ["mapAbility20bp","mapAbility35bp","scoreSegDup","bStatistic","cDNApos","CDSpos","Dst2Splice","GerpN","GerpS","mamPhCons","mamPhyloP","minDistTSE","minDistTSS","priPhCons","priPhyloP","protpos","relcDNApos","relCDSpos","relProtpos","verPhCons","verPhyloP","mutIndex","dnaHelT","dnaMGW","dnaProT","dnaRoll"] # ,"fitCons"
numEncLst = ["EncExp","EncH3K27Ac","EncH3K4Me1","EncH3K4Me3","EncNucleo","EncOCC","EncOCCombPVal","EncOCDNasePVal","EncOCFairePVal","EncOCpolIIPVal","EncOCctcfPVal","EncOCmycPVal","EncOCDNaseSig","EncOCFaireSig","EncOCpolIISig","EncOCctcfSig","EncOCmycSig","EncTFBS","EncTFBSPeaks","EncTFBSPeaksMax"]

parser = OptionParser("%prog [options]")
parser.add_option("-m", "--model", dest="model", help="Model (def input.model)",default="input.model")
parser.add_option("-s", "--score", dest="score", help="Only report score as output",default=False,action="store_true")
parser.add_option("--hcdiff",dest="hcdiff", help="Reverse time orientation of data in conversion",default=False,action="store_true")
parser.add_option("--noGC",dest="noGC", help="Model does not use GC/CpG values",default=False,action="store_true")
parser.add_option("--noSegDup",dest="noSegDup", help="Model does not use SegDup values",default=False,action="store_true")
parser.add_option("--noMap",dest="noMap", help="Model does not use mapabillity values",default=False,action="store_true")
parser.add_option("--noInteractions",dest="noInteractions", help="Model does not use additional interaction terms with Consequence",default=False,action="store_true")
parser.add_option("--noPolyPhen",dest="noPolyPhen", help="Model does not use PolyPhen values",default=False,action="store_true")
parser.add_option("--noEncode",dest="noEncode", help="Model does not use the Encode derived values",default=False,action="store_true")
parser.add_option("--onlyEncode",dest="onlyEncode", help="Models considers only Encode derived values",default=False,action="store_true")
parser.add_option("--EncodeInteraction",dest="EncodeInteraction", help="Extend interactions by Encode features",default=False,action="store_true")
parser.add_option("--quadratic",dest="quadratic", help="Expand product list by quadratic values",default=False,action="store_true")
parser.add_option("--inclTopInteractions", dest="TopInteractions", help="File with interactions showing highest AUC (e.g. interactions/interaction_AUC_Cutoff0.6.txt)",default="")
parser.add_option("--assignment", dest="assignment", help="Print assignment of features",default=False,action="store_true")
parser.add_option("--NA", dest="NA", help="Also print lines where posterior cannot be determined",default=False,action="store_true")
parser.add_option("--ext", dest="ext", help="Expect extended annotation input",default=False,action="store_true")
(options, args) = parser.parse_args()

if options.ext:
  transform = transform_ext+transform_p2
else:
  transform = transform_sm+transform_p2

if options.noGC:
  for ind,(name,value) in enumerate(transform):
    if name == "GC" or name == "CpG":
      transform[ind] = (name,("REMOVE",value[1],value[2]))
  sys.stderr.write('Removing GC/CpG...\n')

if options.noSegDup:
  for ind,(name,value) in enumerate(transform):
    if name == "scoreSegDup":
      transform[ind] = (name,("REMOVE",value[1],value[2]))
  prodlst.remove("scoreSegDup")
  sys.stderr.write('Removing SegDup...\n')

if options.noMap:
  for ind,(name,value) in enumerate(transform):
    if (name == "mapAbility20bp") or (name == "mapAbility35bp"):
      transform[ind] = (name,("REMOVE",value[1],value[2]))
  prodlst.remove("mapAbility20bp")
  prodlst.remove("mapAbility35bp")
  sys.stderr.write('Removing mapability...\n')

if options.noPolyPhen:
  for ind,(name,value) in enumerate(transform):
    if name.startswith("PolyPhen"):
      transform[ind] = (name,("REMOVE",None,None))
  sys.stderr.write('Removing PolyPhen...\n')

if options.noEncode:
  for ind,(name,value) in enumerate(transform):
    if name.startswith("Enc"):
      transform[ind] = (name,("REMOVE",None,None))
  sys.stderr.write('Removing Encode values...\n')

if options.onlyEncode:
  for ind,(name,value) in enumerate(transform):
    if not name.startswith("Enc"):
      transform[ind] = (name,("REMOVE",None,None))
  sys.stderr.write('Keeping only Encode-derived values...\n')

if options.hcdiff:
  sys.stderr.write('Changing time orientation of events...\n')

if options.EncodeInteraction:
  sys.stderr.write('Extending product list...\n')
  prodlst.extend(numEncLst)
  
catvals = {}
intvals = {}
indicators = {}
products = {}
AApairs = {}
BasePairs = {}
SplicePairs = {}
InterPairs = {}

ind = 1
label_assignment = []
for key,(operation,value,capped) in transform:
  if operation != "REMOVE":
    if key in categorical:
      for category in categorical[key]:
        if key not in catvals: catvals[key] = [(category,ind)]
        else: catvals[key].append((category,ind))
        label_assignment.append((ind,(key,category)))
        ind += 1
    else:
      intvals[key] = ind
      label_assignment.append((ind,key))
      ind += 1
  if capped == "IND":
    indicators[key] = ind
    label_assignment.append((ind,"IND_"+key))
    ind += 1

products = {}
if not options.noInteractions and not options.onlyEncode:
  for key in categorical["Consequence"]:
    products[key] = {}
    for key2 in prodlst:
      products[key][key2] = ind
      label_assignment.append((ind,(key,key2)))
      ind += 1

if not options.onlyEncode:
  for elem in categorical["Dst2SplType"][:-1]:
    SplicePairs[elem] = ind
    label_assignment.append((ind,(elem,"Dst2Splice")))
    ind += 1
    indicators[elem] = ind
    label_assignment.append((ind,"IND_"+elem))
    ind += 1

if not options.onlyEncode:
  for oB in categorical["Ref"]:
    for nB in categorical["Alt"]:
      BasePairs[(oB,nB)] = ind
      label_assignment.append((ind,(oB,nB)))
      ind += 1

if not options.onlyEncode:
  for oAA in categorical["oAA"][:-1]:
    for nAA in categorical["nAA"][:-1]:
      AApairs[(oAA,nAA)] = ind
      label_assignment.append((ind,(oAA,nAA)))
      ind += 1

cpairs = set(map(lambda (x,y): y ,label_assignment))
if os.path.exists(options.TopInteractions):
  infile = open(options.TopInteractions)
  for line in infile:
    fields = line.rstrip().split('\t')
    if len(fields) == 2:
      if (fields[0] in intvals) and (fields[1] in intvals):
        if ((fields[0],fields[1]) in cpairs) or ((fields[1],fields[0]) in cpairs): continue
        label_assignment.append((str(ind),(fields[0],fields[1])))
        cpairs.add((fields[0],fields[1]))
        InterPairs[(fields[0],fields[1])] = ind
        ind += 1
      elif (fields[0] in categorical) and (fields[1] in intvals):
        if (fields[0] == "Consequence") and (fields[1] in prodlst): continue
        for cat in categorical[fields[0]]:
          if (((fields[0],cat),fields[1]) in cpairs) or ((fields[1],(fields[0],cat)) in cpairs): continue
          label_assignment.append((str(ind),((fields[0],cat),fields[1])))
          cpairs.add(((fields[0],cat),fields[1]))
          InterPairs[((fields[0],cat),fields[1])] = ind
          ind += 1
      elif (fields[0] in intvals) and (fields[1] in categorical):
        if (fields[1] == "Consequence") and (fields[0] in prodlst): continue
        for cat in categorical[fields[1]]:
          if (((fields[1],cat),fields[0]) in cpairs) or ((fields[0],(fields[1],cat)) in cpairs): continue
          label_assignment.append((str(ind),((fields[1],cat),fields[0])))
          cpairs.add(((fields[1],cat),fields[0]))
          InterPairs[((fields[1],cat),fields[0])] = ind
          ind += 1
      elif (fields[0] in categorical) and (fields[1] in categorical):
        for cat1 in categorical[fields[0]]:
          for cat2 in categorical[fields[1]]:
            if (((fields[0],cat1),(fields[1],cat2)) in cpairs) or (((fields[1],cat2),(fields[0],cat1)) in cpairs): continue
            label_assignment.append((str(ind),((fields[0],cat1),(fields[1],cat2))))
            cpairs.add(((fields[0],cat1),(fields[1],cat2)))
            InterPairs[((fields[0],cat1),(fields[1],cat2))] = ind
            ind += 1
      else:  
        sys.stderr.write("Can not resolve pair (%s,%s)\n"%(fields[0],fields[1]))
  infile.close()

quadratic = {}
quadratic["1"] = {}
if options.quadratic:
  sys.stderr.write('Adding quadratic features...\n')
  for key2 in prodlst:
    quadratic["1"][key2] = ind
    label_assignment.append((str(ind),key2+"**2"))
    ind += 1
  for key in categorical["Consequence"]:
    quadratic[key] = {}
    for key2 in prodlst:
      quadratic[key][key2] = ind
      label_assignment.append((str(ind),(key,key2+"**2")))
      ind += 1

if options.assignment:
  for key,value in label_assignment:
    print key,value
  sys.exit()

factors = [0]
infile = open(options.model)
for line in infile:
  factors.append(float(line))
infile.close()

#chrom,ref,anc,alt,ctype,is_derived,consequence,cons_score,cons_detail,oaa,naa = None,None,None,None,None,None,None,None,None,None,None
#for ind,(name,(operation,value,capped)) in enumerate(transform):
  #if name == "Chrom": chrom = ind
  #elif name == "Ref": ref = ind
  #elif name == "Anc": anc = ind 
  #elif name == "Alt": alt = ind
  #elif name == "Type": ctype = ind
  #elif name == "isDerived": is_derived = ind
  #elif name == "Consequence": consequence = ind
  #elif name == "ConsScore": cons_score = ind
  #elif name == "ConsDetail": cons_detail = ind
  #elif name == "oAA": oaa = ind
  #elif name == "nAA": naa = ind

countlines = 0
transformNames = map(lambda (x,y):x,transform)

for line in sys.stdin:
  countlines += 1
  if line.startswith('#') or line.startswith('Chrom'): 
    if not options.score:
      sys.stdout.write(line.rstrip()+"\tRawScore\n")
    continue
  fields = line.rstrip().split('\t')
  if len(fields) == len(transform):
    #try:
    if True:
      new_fields=[]
      new_names=[]
      skip = False
      score = factors[-1]
      #new_names.append('off')
      #new_fields.append(factors[-1])
      fieldsDict = dict(zip(transformNames,fields))
  
      oAA,nAA = fieldsDict['oAA'],fieldsDict['nAA']
      Ref,Alt = fieldsDict['Ref'],fieldsDict['Alt']
      vtype = fieldsDict['Type']

      if options.hcdiff and Ref != fieldsDict["Anc"]:
        Ref,Alt = fieldsDict['Alt'],fieldsDict['Ref']
        fieldsDict['Alt'],fieldsDict['Ref'] = Ref,Alt
        oAA,nAA = fieldsDict['nAA'],fieldsDict['oAA']
        fieldsDict['nAA'],fieldsDict['oAA'] = oAA,nAA
        if "STOP_GAINED" == fieldsDict["Consequence"]:
          fieldsDict["Consequence"] = "STOP_LOST"
        elif "STOP_LOST" == fieldsDict["Consequence"]:
          fieldsDict["Consequence"] = "STOP_GAINED"
        if "INS" == fieldsDict["Type"]:
          vtype = "DEL"
          fieldsDict["Type"] = vtype
        elif "DEL" == fieldsDict["Type"]:
          vtype = "INS"
          fieldsDict["Type"] = vtype
      
      if (Ref not in validbases) or (vtype != "SNV"): 
        Ref = "N"
        fieldsDict['Ref'] = Ref
      if (Alt not in validbases) or (vtype != "SNV"): 
        Alt = "N"
        fieldsDict['Alt'] = Alt

      if (fieldsDict['oAA'] not in validaminoacids) or (vtype != "SNV"): 
        oAA = "unknown"
        fieldsDict['oAA'] = "unknown"
      if (fieldsDict['nAA'] not in validaminoacids) or (vtype != "SNV"): 
        nAA = "unknown"
        fieldsDict['nAA'] = "unknown"
      
      if (fieldsDict["Length"] != "NA") and (int(fieldsDict["Length"]) >= 50): 
        fieldsDict["Length"]="49"

      # OPERATION: TV
      if fieldsDict['isTv'] == "NA": 
        if (Ref,Alt) in transitions: fieldsDict['isTv'] = "0"
        elif (Ref,Alt) in transversions: fieldsDict['isTv'] = "1"
        else: fieldsDict['isTv'] = "0.5"

      # OPERATION: SPECIAL
      if fieldsDict['scoreSegDup'] == "NA":
        fieldsDict['scoreSegDup'] = "0"
      else:
        fieldsDict['scoreSegDup'] = "1"

      # OPERATION: SET
      if (fieldsDict['Consequence'] not in consequence_set):
          fieldsDict['Consequence'] = "UNKNOWN"

      for ind,(name,(operation,value,capped)) in enumerate(transform):
        if (operation == "REMOVE"): continue
        if (operation == "DROP") and (fieldsDict[name] == "NA"):
          skip = True
          break

        if operation == "LOG":
          if fieldsDict[name] == "NA": # No value observed
            fieldsDict[name] = value
          if float(fieldsDict[name]) == 0: # No log transformable value observed
            if capped != None:
              fieldsDict[name] = "%.4f"%(math.log(float(capped)))
            else:
              fieldsDict[name] = "0"
          else:
            fieldsDict[name] = "%.4f"%(math.log(float(fieldsDict[name])))
        elif fieldsDict[name] == "NA":
          if name in categorical: # Categorical features
            if operation == "NS":
              if vtype == "SNV" and fieldsDict['Consequence'] == "NON_SYNONYMOUS": 
                fieldsDict[name] = value
              else:
                fieldsDict[name] = "UD"
            elif operation == None and value != None:
              fieldsDict[name] = value
            #else:
              #sys.stderr.write('Should not happen [1]: %s %s %s %s!\n'%(name,operation,str(value),capped))
              #sys.exit()
          else: # Numerical features
            if operation == "NS":
              if vtype == "SNV" and fieldsDict['Consequence'] == "NON_SYNONYMOUS":
                fieldsDict[name] = value
              else:
                fieldsDict[name] = "0"
            elif operation == None and value != None: 
              fieldsDict[name] = value
            elif operation == "0":
              fieldsDict[name] = "0"
            #else:
              #sys.stderr.write('Should not happen [2]: %s (%s) %s %s %s!\n'%(name,fieldsDict[name],operation,str(value),capped))
              #sys.exit()
        elif fieldsDict[name] == "TRUE": 
          fieldsDict[name] = "1"
        elif fieldsDict[name] == "FALSE": 
          fieldsDict[name] = "0"
        elif operation == "0" and float(fieldsDict[name]) > value:
          fieldsDict[name] = capped
        #else:
          #if (name not in categorical) and fieldsDict[name] == "NA":
            #sys.stderr.write('Should not happen [3]: %s (%s) %s %s %s!\n'%(name,fieldsDict[name],operation,str(value),capped))
            #sys.exit()

        if name in categorical: # Categorical features
          for cat,sind in catvals[name]:
            fieldsDict[(name,cat)] = "0"
            if cat == fieldsDict[name]: 
              score += factors[sind]
              #new_fields.append(factors[sind])
              #new_names.append('%s:%s'%(name,cat))
              fieldsDict[(name,cat)] = "1"
        else: # Numerical features
          if capped == "IND" and (fieldsDict[name] == "0") and (fields[ind] != fieldsDict[name]): 
            score += factors[indicators[name]]
            #new_fields.append(factors[indicators[name]])
            #new_names.append('IND:%s'%(name))
          else:
            score += factors[intvals[name]]*float(fieldsDict[name])
            #new_fields.append(factors[intvals[name]]*float(fieldsDict[name]))
            #new_names.append('%s'%(name))

      if not skip:
        if not options.onlyEncode:
          if fieldsDict["Consequence"] in products:
            for key2,pind in products[fieldsDict["Consequence"]].iteritems()  :
              hval2 = fieldsDict.get(key2,None)
              if hval2 != None: 
                score += factors[pind]*float(hval2)
                #new_fields.append(factors[pind]*float(hval2))
                #new_names.append('Consequence(x)%s'%key2)

          for key2,pind in quadratic["1"].iteritems():
            hval2 = cvalues.get(key2,None)
            if hval2 != None and float(hval2) != 0: score += factors[pind]*float(hval2)**2.0
            #new_fields.append(factors[pind]*float(hval2)**2)
            #new_names.append('%s**2'%key2)
          if fieldsDict["Consequence"] in quadratic:
            for key2,pind in quadratic[fieldsDict["Consequence"]].iteritems():
              hval2 = cvalues.get(key2,None)
              if hval2 != None and float(hval2) != 0: score += factors[pind]*float(hval2)**2.0
              #new_fields.append(factors[pind]*float(hval2)**2)
              #new_names.append('Consequence(x)%s**2'%key2)
      
          if fieldsDict["Dst2SplType"] != "unknown" and fieldsDict["Dst2Splice"] != "NA": 
            score += factors[SplicePairs[fieldsDict["Dst2SplType"]]]*float(fieldsDict["Dst2Splice"])
            #new_fields.append(factors[SplicePairs[fieldsDict["Dst2SplType"]]]*float(fieldsDict["Dst2Splice"]))
            #new_names.append('Dst2SplType(x)Dst2Splice')
          
          for splType in ["DONOR","ACCEPTOR"]:
            if fieldsDict["Dst2SplType"] != splType: 
              score += factors[indicators[splType]]
              #new_fields.append(factors[indicators[splType]])
              #new_names.append('IND:%s'%splType)
            
          score += factors[BasePairs[Ref,Alt]]
          #new_fields.append(factors[BasePairs[Ref,Alt]])
          #new_names.append('REF(x)ALT')
        
          if oAA != "unknown" and nAA != "unknown": 
            score += factors[AApairs[oAA,nAA]]
            #new_fields.append(factors[AApairs[oAA,nAA]])
            #new_names.append('oAA(x)nAA')

          for (key1,key2),pind in InterPairs.iteritems():
            score += factors[pind]*float(fieldsDict[key1])*float(fieldsDict[key2])
            #new_fields.append(factors[pind]*float(fieldsDict[key1])*float(fieldsDict[key2]))
            #new_names.append('%s(x)%s'%(key1,key2))
        
        if options.score:
          sys.stdout.write("%.20f"%(score)+"\n")
          #sys.stderr.write(" ".join(map(lambda (x,y):"%s:%.4f"%(x,y),filter(lambda (x,y): y!=0,zip(new_names,new_fields))))+"\n")
        else:
          sys.stdout.write(line.rstrip()+"\t%.6f"%(score)+"\n")

    #except:
      #sys.stderr.write('Unexpected error scoring line %d:\n'%(countlines))
      #sys.stderr.write(str(fields)+"\n")
  else:
    sys.stderr.write('Unexpected number of fields in line %d: T:%d F:%d\n'%(countlines,len(transform),len(fields)))
    sys.stderr.write(str(transform)+"\n")
    sys.stderr.write(str(fields)+"\n")
    sys.exit()
